# -*- coding: utf-8 -*-

"""
Integration part for hooking into distutils/setuptools

Rationale:
The ``use_pyscaffold`` keyword is unknown to setuptools' setup(...) command,
therefore the ``entry_points`` are checked for a function to handle this
keyword which is ``pyscaffold_keyword`` below. This is where we hook into
setuptools and apply the magic of setuptools_scm and pbr.
"""

from __future__ import division, print_function, absolute_import

import os
from distutils.cmd import Command

from pyscaffold.contrib import pbr_read_setup_cfg, scm_get_version
from pyscaffold.utils import check_setuptools_version
from pyscaffold.repo import get_git_root
from pyscaffold.pytest_runner import PyTest

__author__ = "Florian Wilhelm"
__copyright__ = "Blue Yonder"
__license__ = "new BSD"


def version2str(version):
    """Creates a PEP440 version string

    Args:
        version (:obj:`setuptools_scm.version.ScmVersion`): version object

    Returns:
        str: version string
    """
    if version.exact or not version.distance > 0:
        return version.format_with('{tag}')
    else:
        distance = version.distance
        version = str(version.tag)
        if '.dev' in version:
            version, tail = version.rsplit('.dev', 1)
            assert tail == '0', 'own dev numbers are unsupported'
        return '{}.post0.dev{}'.format(version, distance)


def local_version2str(version):
    """Create the local part of a PEP440 version string

    Args:
        version (:obj:`setuptools_scm.version.ScmVersion`): version object

    Returns:
        str: local version
    """
    if version.exact:
        return ''
    else:
        if version.dirty:
            return version.format_with('+n{node}.dirty')
        else:
            return version.format_with('+n{node}')


def deactivate_pbr_authors_changelog():
    """Deactivate automatic generation of AUTHORS and ChangeLog file

    This is an automatism of pbr and we rather keep track of our own
    AUTHORS.rst and CHANGES.rst files.
    """
    os.environ['SKIP_GENERATE_AUTHORS'] = "1"
    os.environ['SKIP_WRITE_GIT_CHANGELOG'] = "1"


def build_cmd_docs():
    """Return Sphinx's BuildDoc if available otherwise a dummy command

    Returns:
        :obj:`~distutils.cmd.Command`: command object
    """
    try:
        from sphinx.setup_command import BuildDoc
    except ImportError:
        class NoSphinx(Command):
            user_options = []

            def initialize_options(self):
                raise RuntimeError("Sphinx documentation is not installed, "
                                   "run: pip install sphinx")

        return NoSphinx
    else:
        return BuildDoc


def pyscaffold_keyword(dist, keyword, value):
    """Handles the `use_pyscaffold` keyword of the setup(...) command

    Args:
        dist (:obj:`setuptools.dist`): distribution object as
        keyword (str): keyword argument = 'use_pyscaffold'
        value: value of the keyword argument
    """
    check_setuptools_version()
    if value:
        # If value is a dictionary we keep it otherwise use for configuration
        value = value if isinstance(value, dict) else dict()
        value.setdefault('root', get_git_root(default='.'))
        value.setdefault('version_scheme', version2str)
        value.setdefault('local_scheme', local_version2str)
        if os.path.exists('PKG-INFO'):
            value.pop('root', None)
        command_options = dist.command_options.copy()
        cmdclass = dist.cmdclass.copy()
        deactivate_pbr_authors_changelog()
        pbr_read_setup_cfg(dist, keyword, True)
        dist.metadata.version = scm_get_version(**value)
        # Adding old command classes and options since pbr seems to drop these
        dist.cmdclass['doctest'] = build_cmd_docs()
        dist.command_options['doctest'] = {'builder': ('setup.py', 'doctest')}
        dist.cmdclass['test'] = PyTest
        dist.cmdclass.update(cmdclass)
        dist.cmdclass.update(command_options)



import pytest
import sys
import pkg_resources
from setuptools_scm import dump_version, get_version, PRETEND_KEY
from setuptools_scm.version import guess_next_version, meta, format_version
from setuptools_scm.utils import has_command

PY3 = sys.version_info > (2,)


class MockTime(object):
    def __format__(self, *k):
        return 'time'


@pytest.mark.parametrize('tag, expected', [
    ('1.1', '1.2.dev0'),
    ('1.2.dev', '1.2.dev0'),
    ('1.1a2', '1.1a3.dev0'),
    ('23.24.post2+deadbeef', '23.24.post3.dev0'),
    ])
def test_next_tag(tag, expected):
    version = pkg_resources.parse_version(tag)
    assert guess_next_version(version, 0) == expected


VERSIONS = {
    'exact': meta('1.1', None, False),
    'zerodistance': meta('1.1', 0, False),
    'dirty': meta('1.1', None, True),
    'distance': meta('1.1', 3, False),
    'distancedirty': meta('1.1', 3, True),
}


@pytest.mark.parametrize('version,scheme,expected', [
    ('exact', 'guess-next-dev node-and-date', '1.1'),
    ('zerodistance', 'guess-next-dev node-and-date', '1.2.dev0'),
    ('dirty', 'guess-next-dev node-and-date', '1.2.dev0+dtime'),
    ('distance', 'guess-next-dev node-and-date', '1.2.dev3'),
    ('distancedirty', 'guess-next-dev node-and-date', '1.2.dev3+dtime'),
    ('exact', 'post-release node-and-date', '1.1'),
    ('zerodistance', 'post-release node-and-date', '1.1.post0'),
    ('dirty', 'post-release node-and-date', '1.1.post0+dtime'),
    ('distance', 'post-release node-and-date', '1.1.post3'),
    ('distancedirty', 'post-release node-and-date', '1.1.post3+dtime'),
])
def test_format_version(version, monkeypatch, scheme, expected):
    version = VERSIONS[version]
    monkeypatch.setattr(version, 'time', MockTime())
    vs, ls = scheme.split()
    assert format_version(
        version,
        version_scheme=vs,
        local_scheme=ls) == expected


def test_dump_version_doesnt_bail_on_value_error(tmpdir):
    write_to = "VERSION"
    version = str(VERSIONS['exact'].tag)
    with pytest.raises(ValueError) as exc_info:
        dump_version(tmpdir.strpath, version, write_to)
    assert str(exc_info.value).startswith("bad file format:")


def test_dump_version_works_with_pretend(tmpdir, monkeypatch):
    monkeypatch.setenv(PRETEND_KEY, '1.0')
    get_version(write_to=str(tmpdir.join('VERSION.txt')))
    assert tmpdir.join('VERSION.txt').read() == '1.0'


def test_has_command(recwarn):
    assert not has_command('yadayada_setuptools_aint_ne')
    msg = recwarn.pop()
    assert 'yadayada' in str(msg.message)



# -*- coding: utf-8 -*-

"""
This module provides a test runner for setup.py copied over from
https://bitbucket.org/pytest-dev/pytest-runner/
in order to make some improvements.

This file is MIT licensed:

Copyright (c) 2011 Jason R. Coombs <jaraco@jaraco.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import division, print_function, absolute_import

import sys
import os
import shlex

import pkg_resources
from setuptools.command.test import test as TestCommand

__author__ = "Jason R. Coombs, Florian Wilhelm"
__copyright__ = "Jason R. Coombs"
__license__ = "MIT"


class PyTest(TestCommand):
    user_options = [
        ('extras', None, "Install (all) setuptools extras when running tests"),
        ('index-url=', None, "Specify an index url from which to retrieve "
            "dependencies"),
        ('allow-hosts=', None, "Whitelist of comma-separated hosts to allow "
            "when retrieving dependencies"),
        ('addopts=', 'a', "Additional options to be passed verbatim to the "
            "pytest runner")
    ]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.extras = False
        self.index_url = None
        self.allow_hosts = None
        self.addopts = []

    def finalize_options(self):
        if self.addopts:
            self.addopts = shlex.split(self.addopts)

    @staticmethod
    def marker_passes(marker):
        """
        Given an environment marker, return True if the marker is valid
        and matches this environment.
        """
        return (
            marker
            and not pkg_resources.invalid_marker(marker)
            and pkg_resources.evaluate_marker(marker)
        )

    def run(self):
        """
        Override run to ensure requirements are available in this session (but
        don't install them anywhere).
        """
        self._build_egg_fetcher()
        if self.distribution.install_requires:
            self.distribution.fetch_build_eggs(
                self.distribution.install_requires)
        if self.distribution.tests_require:
            self.distribution.fetch_build_eggs(self.distribution.tests_require)
        extras_require = self.distribution.extras_require or {}
        for spec, reqs in extras_require.items():
            name, sep, marker = spec.partition(':')
            if marker and not self.marker_passes(marker):
                continue
            # always include unnamed extras
            if not name or self.extras:
                self.distribution.fetch_build_eggs(reqs)
        if self.dry_run:
            self.announce('skipping tests (dry run)')
            return
        self.with_project_on_sys_path(self.run_tests)
        if self.result_code:
            raise SystemExit(self.result_code)
        return self.result_code

    def _build_egg_fetcher(self):
        """Build an egg fetcher that respects index_url and allow_hosts"""
        # modified from setuptools.dist:Distribution.fetch_build_egg
        from setuptools.command.easy_install import easy_install
        main_dist = self.distribution
        # construct a fake distribution to store the args for easy_install
        dist = main_dist.__class__({'script_args': ['easy_install']})
        dist.parse_config_files()
        opts = dist.get_option_dict('easy_install')
        keep = (
            'find_links', 'site_dirs', 'index_url', 'optimize',
            'site_dirs', 'allow_hosts'
        )
        for key in opts.keys():
            if key not in keep:
                del opts[key]   # don't use any other settings
        if main_dist.dependency_links:
            links = main_dist.dependency_links[:]
            if 'find_links' in opts:
                links = opts['find_links'][1].split() + links
            opts['find_links'] = ('setup', links)
        if self.allow_hosts:
            opts['allow_hosts'] = ('test', self.allow_hosts)
        if self.index_url:
            opts['index_url'] = ('test', self.index_url)
        install_dir_func = getattr(dist, 'get_egg_cache_dir', os.getcwd)
        install_dir = install_dir_func()
        cmd = easy_install(
            dist, args=["x"], install_dir=install_dir, exclude_scripts=True,
            always_copy=False, build_directory=None, editable=False,
            upgrade=False, multi_version=True, no_report=True
        )
        cmd.ensure_finalized()
        main_dist._egg_fetcher = cmd

    def run_tests(self):
        try:
            import pytest
        except ImportError:
            raise RuntimeError("PyTest is not installed, run: "
                               "pip install pytest pytest-cov")
        errno = pytest.main(self.addopts)
        sys.exit(errno)



import numpy as np
import scipy
from numba import jit

__all__ = ['SimpleCycleReservoir']


class SimpleCycleReservoir:

    def __init__(self, n_nodes=30, regularization=1e-8, cyclic_weight=0.5, input_weight=0.5, random_seed=123):
        # Save attributes
        self.n_nodes = np.int32(np.round(n_nodes))
        self.regularization = np.float32(regularization)
        self.cyclic_weight = np.float32(cyclic_weight)
        self.input_weight = np.float32(input_weight)
        self.seed = random_seed

        # Generate reservoir
        self.generate_reservoir()

    def generate_reservoir(self):
        """Generates transition weights"""
        # Set reservoir weights
        self.weights = np.zeros((self.n_nodes, self.n_nodes), dtype=np.float32)
        self.weights[0, -1] = self.cyclic_weight
        for i in range(self.n_nodes - 1):
            self.weights[i + 1, i] = self.cyclic_weight

        # Set out to none to indicate untrained ESN
        self.out_weights = None

    def draw_reservoir(self):
        """Vizualizes reservoir.

        Note: Requires 'networkx' package.

        """
        import networkx as nx
        graph = nx.DiGraph(self.weights)
        nx.draw(graph)

    @jit(cache=True)
    def generate_states(self, x, burn_in=30):
        """Generates states given some column vector x"""
        # Initialize new random state
        random_state = np.random.RandomState(self.seed)

        # Set and scale input weights (for memory length and non-linearity)
        self.in_weights = np.full(shape=(x.shape[1], self.n_nodes), fill_value=self.input_weight, dtype=np.float32)
        self.in_weights *= np.sign(random_state.uniform(low=-1.0, high=1.0, size=self.in_weights.shape))

        # Generate with jit version
        return generate_states_inner_loop(x, self.n_nodes, self.in_weights, self.weights, burn_in)

    @jit(cache=True)
    def train(self, y, x, burn_in=30):
        """Trains the Echo State Network.

        Trains the out weights on the random network. This is needed before being able to make predictions.
        Consider running a burn-in of a sizable length. This makes sure the state  matrix has converged to a
        'credible' value.

        Parameters
        ----------
        y : array
            Column vector of y values
        x : array or None
            Optional matrix of inputs (features by column)
        burn_in : int
            Number of inital time steps to be discarded for model inference

        Returns
        -------
        state, y, burn_in : tuple
            Returns the state, the y values provided and the number of time steps used for burn_in.
            These data can be used for diagnostic purposes (e.g. vizualization of activations), or for
            cross-validation.

        """
        # Set types
        if not (y.dtype == np.float32 and x.dtype == np.float32):
            x = x.astype(np.float32)
            y = y.astype(np.float32)

        burn_in = np.int32(burn_in)

        # Get states
        state = self.generate_states(x, burn_in=burn_in)

        # Concatenate inputs with node states
        train_x = state
        train_y = y[burn_in:]  # Include everything after burn_in

        # Ridge regression
        ridge_x = train_x.T @ train_x + self.regularization * np.eye(train_x.shape[1], dtype=np.float32)
        ridge_y = train_x.T @ train_y

        # Solve for out weights
        # try:
        # Cholesky solution (fast)
        self.out_weights = np.linalg.solve(ridge_x, ridge_y).reshape(-1, 1)
        # except np.linalg.LinAlgError:
        #     # Pseudo-inverse solution (robust solution if ridge_x is singular)
        #     self.out_weights = (scipy.linalg.pinvh(ridge_x) @ ridge_y).reshape(-1, 1)

        # Return all data for computation or visualization purposes
        return state, y, burn_in

    @jit(cache=True)
    def validation_score(self, y, x, folds=5, scoring_method='L2', burn_in=30):
        """Trains and gives k-folds validation score"""
        # Set types
        x = x.astype(np.float32, copy=False)
        y = y.astype(np.float32, copy=False)

        # Get states
        state = self.generate_states(x, burn_in=np.int32(burn_in))

        # Placeholder
        scores = np.zeros(folds, dtype=np.float32)

        # Get y
        y = y[burn_in:]

        # Fold size
        fold_size = y.shape[0] // folds

        for k in range(folds):

            # Validation folds
            start_index = k * fold_size
            stop_index = start_index + fold_size

            # Indices
            validation_indices = np.arange(start_index, stop_index, dtype=np.int32)

            # Train mask
            train_mask = np.ones(y.shape[0], dtype=bool)
            train_mask[validation_indices] = False

            # Concatenate inputs with node states
            train_x = state[train_mask]
            train_y = y[train_mask]

            # Ridge regression
            ridge_x = train_x.T @ train_x + self.regularization * np.eye(train_x.shape[1])
            ridge_y = train_x.T @ train_y

            # Solve for out weights
            # try:
            # Cholesky solution (fast)
            out_weights = np.linalg.solve(ridge_x, ridge_y).reshape(-1, 1)
            # except np.linalg.LinAlgError:
            #     # Pseudo-inverse solution
            #     out_weights = (scipy.linalg.pinvh(ridge_x) @
            #                    ridge_y).reshape(-1, 1)  # Robust solution if ridge_x is singular

            # Validation set
            validation_x = state[validation_indices]
            validation_y = y[validation_indices]

            # Predict
            prediction = validation_x @ out_weights

            # Score
            scores[k] = self.error(prediction, validation_y, scoring_method)

        # Return mean validation score
        return scores.mean()

    @jit(cache=True)
    def train_validate_multiple(self,
                                y,
                                x,
                                series_weights,
                                folds=5,
                                scoring_method='L2',
                                full_train=False,
                                skip_folds=False,
                                burn_in=30):
        """Trains and gives k-folds validation score for multiple series

        Parameters
        ----------
        y : array
            Multiple output series
        x : array
            Multiple input series, belonging only to the series in y with the same index
        series_weights : array
            Array of n_series values, with the weighting of the series toward the regression
        full_train : bool
            Will train on full training set
        skip_folds : bool
            Skips k-folds if enabled

        Returns
        -------
        scores : array
            Errors for all series during cross-validation on weighted ridge regression

        """
        # Checks
        assert y.shape == x.shape, 'Data matrices not of equal shape'
        if not y.dtype == np.float32:
            y = y.astype(np.float32)

        if not x.dtype == np.float32:
            x = x.astype(np.float32)

        burn_in = np.int32(burn_in)

        # Easy retrieval
        t_steps = y.shape[0]
        n_series = y.shape[1]
        effective_length = t_steps - burn_in
        samples = effective_length * n_series

        # Concatenate all states
        states = np.zeros((samples, self.n_nodes + 1), dtype=np.float32)  # Add one column for intercept
        all_y = np.zeros((samples, 1), dtype=np.float32)
        for n, start_index in enumerate(range(0, samples, effective_length)):

            # Get states
            states[start_index:start_index + effective_length, :] = self.generate_states(x[:, n].reshape(-1, 1),
                                                                                         burn_in=burn_in)

            # Concatenate output
            all_y[start_index:start_index + effective_length, 0] = y[burn_in:, n]

        # Series weights
        sample_weights = series_weights.repeat(effective_length)

        # Placeholders
        scores = np.full(folds, np.nan, dtype=np.float32)
        readouts = np.full((self.n_nodes + 1, folds), np.nan, dtype=np.float32)

        if not skip_folds:
            # Shuffle data
            random_state = np.random.RandomState(self.seed + 2)
            permutation = np.arange(samples, dtype=np.int32)
            random_state.shuffle(permutation)
            shuffled_states = states[permutation]
            shuffled_y = all_y[permutation]
            shuffled_weights = sample_weights[permutation]

            # Fold size
            fold_size = samples // folds

            # K-folds
            for k in range(folds):

                # Validation folds
                start_index = k * fold_size
                stop_index = start_index + fold_size

                # Indices
                validation_indices = np.arange(start_index, stop_index, dtype=np.int32)

                # Train mask
                train_mask = np.ones(samples, dtype=bool)
                train_mask[validation_indices] = False

                # Concatenate inputs with node states
                train_x = shuffled_states[train_mask]
                train_y = shuffled_y[train_mask]
                masked_weights = shuffled_weights[train_mask]

                # Center weights at 1 so to not distort L2 regularizion
                w = (masked_weights / masked_weights.mean()).reshape(-1, 1)  # Make column vector

                # Weighted Ridge regression
                ridge_x = train_x.T @ (w * train_x) + self.regularization * np.eye(train_x.shape[1])
                ridge_y = train_x.T @ (w * train_y)

                # Solve for out weights
                # try:
                # Cholesky solution (fast)
                out_weights = np.linalg.solve(ridge_x, ridge_y).reshape(-1, 1)
                # except np.linalg.LinAlgError:
                #     # Pseudo-inverse solution (robust solution if ridge_x is singular)
                #     out_weights = (scipy.linalg.pinvh(ridge_x) @ ridge_y).reshape(-1, 1)

                # Validation set
                validation_x = shuffled_states[validation_indices]
                validation_y = shuffled_y[validation_indices]

                # Predict
                prediction = validation_x @ out_weights

                # Save
                scores[k] = self.error(prediction, validation_y, scoring_method)
                readouts[:, k] = out_weights.reshape(-1,)

        # Do a train on the full x and y dataset if demanded
        if full_train:
            # Rename by convention
            train_x = states
            train_y = all_y

            # Center weights at 1 so to not distort L2 regularizion
            w = (sample_weights / sample_weights.mean()).reshape(-1, 1)  # Make column vector

            # Weighted Ridge regression
            ridge_x = train_x.T @ (w * train_x) + self.regularization * np.eye(train_x.shape[1])
            ridge_y = train_x.T @ (w * train_y)

            # Solve for out weights
            self.out_weights = np.linalg.solve(ridge_x, ridge_y).reshape(-1, 1)
        else:
            # Set weights to best in CV
            best_index = np.argmin(scores)
            self.out_weights = readouts[:, best_index]

        # Return validation scores
        return scores

    def test(self, y, x, out_weights=None, scoring_method='L2', burn_in=30, alpha=1., **kwargs):
        """Tests and scores against known output.

        Parameters
        ----------
        y : array
            Column vector of known outputs
        x : array or None
            Any inputs if required
        scoring_method : {'L2', 'mse', 'rmse', 'nrmse', 'tanh'}
            Evaluation metric used to calculate error
        burn_in : int
            Number of time steps to exclude from prediction initially
        alpha : float
            Alpha coefficient to scale the tanh error transformation: alpha * tanh{(1 / alpha) * error}

        Returns
        -------
        error : float
            Error between prediction and known outputs

        """
        if not y.dtype == np.float32:
            y = y.astype(np.float32)

        if not x.dtype == np.float32:
            x = x.astype(np.float32)

        # Run prediction
        y_predicted = self.predict_stepwise(x, out_weights=out_weights)

        # Checks
        assert y_predicted.shape[0] == y.shape[0]

        # Return error
        return self.error(y_predicted[burn_in:], y[burn_in:], scoring_method, alpha=alpha)

    def predict_stepwise(self, x, out_weights=None, **kwargs):
        """Predicts a specified number of steps into the future for every time point in y-values array.

        Parameters
        ----------
        x : numpy array or None
            Prediction inputs
        out_weights : numpy array (2D column vector)
            The weights to use for prediction. Overrides any trained weights stored on the object.

        Returns
        -------
        y_predicted : numpy array
            Array of predictions at every time step of shape (times, steps_ahead)

        """
        # Check if ESN has been trained
        if self.out_weights is None and out_weights is None:
            raise ValueError('Error: Train model or provide out_weights')

        # Get states
        state = self.generate_states(x, burn_in=np.int32(0))

        # Select weights
        if not out_weights is None:
            weights = out_weights  # Provided
        else:
            weights = self.out_weights  # From training

        # Predict
        y_predicted = state @ weights

        # Return predictions
        return y_predicted

    def error(self, predicted, target, method='L2', alpha=1.):
        """Evaluates the error between predictions and target values.

        Parameters
        ----------
        predicted : array
            Predicted value
        target : array
            Target values
        method : {'L2', 'mse', 'tanh', 'rmse', 'nmse', 'nrmse', 'tanh-nmse', 'log-tanh', 'log'}
            Evaluation metric. 'tanh' takes the hyperbolic tangent of mse to bound its domain to [0, 1] to ensure
            continuity for unstable models. 'log' takes the logged mse, and 'log-tanh' takes the log of the squeezed
            normalized mse. The log ensures that any variance in the GP stays within bounds as errors go toward 0.
        alpha : float
            Alpha coefficient to scale the tanh error transformation: alpha * tanh{(1 / alpha) * error}.
            This squeezes errors onto the interval [0, alpha].
            Default is 1. Suggestions for squeezing errors > n * stddev of the original series
            (for tanh-nrmse, this is the point after which difference with y = x is larger than 50%,
             and squeezing kicks in):

        Returns
        -------
        error : float
            The error as evaluated with the metric chosen above

        """
        errors = target.ravel() - predicted.ravel()

        # Adjust for NaN and np.inf in predictions (unstable solution)
        if not np.all(np.isfinite(predicted)):
            # print("Warning: some predicted values are not finite")
            errors = np.inf

        # Compute mean error
        if method == 'L2':
            error = np.linalg.norm(errors, ord=2)  # Technically the same as rmse
        elif method == 'mse':
            error = np.mean(np.square(errors))
        elif method == 'tanh':
            error = alpha * np.tanh(np.mean(np.square(errors)) / alpha)  # To 'squeeze' mse onto the interval (0, 1)
        elif method == 'rmse':
            error = np.sqrt(np.mean(np.square(errors)))
        elif method == 'nmse':
            error = np.mean(np.square(errors)) / np.square(target.ravel().std(ddof=1))
        elif method == 'nrmse':
            error = np.sqrt(np.mean(np.square(errors))) / target.ravel().std(ddof=1)
        elif method == 'tanh-nmse':
            nmse = np.mean(np.square(errors)) / np.square(target.ravel().std(ddof=1))
            error = alpha * np.tanh(nmse / alpha)
        else:
            raise ValueError('Scoring method not recognized')
        return error

@jit(nopython=True, cache=True)
def generate_states_inner_loop(x, n_nodes, in_weights, weights, burn_in):
    # Calculate correct shape
    rows = x.shape[0]

    # Compute states
    state = x @ in_weights

    # Set last state
    previous_state = np.zeros(n_nodes, dtype=np.float32)

    # Train iteratively
    for t in range(rows):
        state[t] += weights @ previous_state
        state[t] = np.tanh(state[t])
        previous_state = state[t]

    # Add intecept
    state = np.hstack((np.ones((rows, 1), dtype=np.float32), state))

    # Return
    return state[burn_in:]



# Based on the source code of GPyOpt/models/gpmodel.py,
# which is licensed under the BSD 3-clause license

import numpy as np
import GPyOpt
import GPy
import copy
import contextlib
import sys
import traceback
from scipy.special import gammaln, digamma
from paramz.domains import _REAL, _POSITIVE


__all__ = ['RobustGPModel']


class RobustGPModel(GPyOpt.models.GPModel):
    """
    General class for handling a Gaussian Process in GPyOpt.

    Adapted from GPModel (Copyright (c) 2016, the GPyOpt Authors, BSD 3 license)

    :param kernel: GPy kernel to use in the GP model.
    :param noise_var: value of the noise variance if known.
    :param exact_feval: whether noiseless evaluations are available.
    :param normalize_Y: normalization of the outputs to the interval [0,1] (default, True).
    :param optimizer: optimizer of the model. Check GPy for details.
    :param max_iters: maximum number of iterations used to optimize the parameters of the model.
    :param verbose: print out the model messages (default, False).

    .. Note:: This model does Maximum likelihood estimation of the hyper-parameters.

    """
    analytical_gradient_prediction = True

    def __init__(self, noise_var=None, log_space=False, exact_feval=False, normalize_Y=True,
                 max_iters=1000, verbose=True, **kwargs):
        super().__init__()
        self.noise_var = noise_var
        self.log_space = log_space
        self.exact_feval = exact_feval
        self.normalize_Y = normalize_Y
        self.max_iters = max_iters
        self.verbose = verbose
        self.model = None

    def _preprocess_data(self, X, Y):
        # Remove non-finite values
        finite_mask = np.isfinite(Y.ravel())  # TODO: Detect outliers that are not infinite
        infinite_indices = np.nonzero(~finite_mask)[0]

        # Replace with mean
        if not self.model is None and infinite_indices.shape[0] > 0:
            X_inf = X[infinite_indices]
            means, stds = self.model.predict(X_inf)
            Y[infinite_indices] = means + 1e-8 * stds
        elif infinite_indices.shape[0] > 0:
            Y = Y[finite_mask]
            X = X[finite_mask]

        # Log if wanted
        if self.log_space:
            Y = np.log(Y + 1e-8)  # Y will be negative when working with (N)MSE

        # Normalize
        if self.normalize_Y:
            Y -= Y.mean()
            std = Y.std()
            Y /= std if std > 0. else 1.

        return X, Y

    def _create_model(self, X, Y):
        """
        Creates the model given some input data X and Y.
        """
        with Suppressor():
            # Kernel and priors
            self.input_dim = X.shape[1]
            kernel = GPy.kern.Matern52(self.input_dim, ARD=True)
            prior = FixedInverseGamma(.001, .001)
            kernel.lengthscale.set_prior(prior)
            kernel.variance.set_prior(prior)

            # Model
            noise_var = 1. if self.noise_var is None else self.noise_var
            self.model = GPy.models.GPRegression(X, Y, kernel=kernel, noise_var=noise_var)
            self.model.likelihood.variance.set_prior(prior)

            # Evaluation constriant
            if self.exact_feval or noise_var < 1e-6:
                self.model.Gaussian_noise.constrain_fixed(1e-6, warning=False)
            else:
                self.model.Gaussian_noise.constrain_positive(warning=False)

    def updateModel(self, X_all, Y_all):
            """
            Updates the model with new observations.
            """
            X, Y = self._preprocess_data(X_all, Y_all)

            if self.model is None:
                self._create_model(X, Y)
            else:
                self.model.set_XY(X, Y)

            # Update model
            self.model.optimize(optimizer='lbfgs', messages=False, max_iters=self.max_iters,
                                ipython_notebook=True, clear_after_finish=True)

    def predict(self, X):
            """
            Predictions with the model. Returns posterior means and standard deviations at X.
            Note that this is different in GPy where the variances are given.
            """
            if X.ndim == 1:
                X = X[None, :]
            m, v = self.model.predict(X)
            v = np.clip(v, 1e-10, np.inf)
            return m, np.sqrt(v)

    def get_fmin(self):
            """
            Returns the location where the posterior mean is takes its minimal value.
            """
            return self.model.predict(self.model.X)[0].min()

    def predict_withGradients(self, X):
            """
            Returns the mean, standard deviation, mean gradient and standard deviation gradient at X.
            """
            if X.ndim == 1:
                X = X[None, :]
            m, v = self.model.predict(X)
            v = np.clip(v, 1e-10, np.inf)
            dmdx, dvdx = self.model.predictive_gradients(X)
            dmdx = dmdx[:, :, 0]
            dsdx = dvdx / (2 * np.sqrt(v))
            return m, np.sqrt(v), dmdx, dsdx

    def copy(self):
            """
            Makes a safe copy of the model.
            """
            return copy.deepcopy(self)

    def get_model_parameters(self):
            """
            Returns a 2D numpy array with the parameters of the model
            """
            return np.atleast_2d(self.model[:])

    def get_model_parameters_names(self):
            """
            Returns a list with the names of the parameters of the model
            """
            return self.model.parameter_names()


# Added to circumvent a GPy bug
class FixedInverseGamma(GPy.priors.Gamma):
    """
    Implementation of the inverse-Gamma probability function, coupled with random variables.
    :param a: shape parameter
    :param b: rate parameter (warning: it's the *inverse* of the scale)
    .. Note:: Bishop 2006 notation is used throughout the code
    """
    domain = _POSITIVE

    def __init__(self, a, b):
        super().__init__(a, b)
        self._a = float(a)
        self._b = float(b)
        self.constant = -gammaln(self._a) + self._a * np.log(self._b)

    def __str__(self):
        return "iGa({:.2g}, {:.2g})".format(self._a, self._b)

    def lnpdf(self, x):
        return self.constant - (self._a + 1) * np.log(x) - self._b / x

    def lnpdf_grad(self, x):
        return -(self._a + 1.) / x + self._b / x ** 2

    def rvs(self, n):
        return 1. / np.random.gamma(scale=1. / self._b, shape=self._a, size=n)


class Suppressor(object):
    """
    Adapted from: https://stackoverflow.com/questions/2828953/ \
        silence-the-stdout-of-a-function-in-python-without-trashing-sys-stdout-and-resto/40054132#40054132

    """

    def __enter__(self):
        self.stdout = sys.stdout
        sys.stdout = self

    def __exit__(self, type_, value, trace):
        sys.stdout = self.stdout
        if type_ is not None:
            # Do normal exception handling
            traceback.print_tb(trace)

    def write(self, x):
        pass



import torch
import pylab as pl
from IPython import display
import matplotlib.pyplot as plt
import torch.optim as optim
from decimal import Decimal
import numpy as np

lam =1
def hamiltonian(x, p, lam = lam):
    return (1/2)*(x**2 + p**2) + lam*x**4/4


def fforce(t, A = 1):
    return A * torch.sin(t)

def no_fforce(t):
    return 0

def no_reg_loss(X , y, ydot, out_weights, force_t = None, 
                reg = True, ode_coefs = None, mean = True,
               enet_strength = None, enet_alpha = None, init_conds = None, lam = 1):
    y, p = y[:,0].view(-1,1), y[:,1].view(-1,1)
    ydot, pdot = ydot[:,0].view(-1,1), ydot[:,1].view(-1,1)
    
    #with paramization
    L =  (ydot - p)**2 + (pdot + y + lam * y**3   - force_t)**2
    
    #if mean:
    L = torch.mean(L)
    #    hi = float(L)
    
    #if reg:
    weight_size_sq = torch.mean(torch.square(out_weights))
    weight_size_L1 = torch.mean(torch.abs(out_weights))
    L_reg = enet_strength*(enet_alpha * weight_size_sq + (1- enet_alpha) * weight_size_L1)
    L = L + 0.1 * L_reg 
    

#     y0, p0 = init_conds
#     ham = hamiltonian(y, p)
#     ham0 = hamiltonian(y0, p0)
#     L_H = (( ham - ham0).pow(2)).mean()
#     assert L_H >0

#     L = L +  L_H
    
    #print("L1", hi, "L_elastic", L_reg, "L_H", L_H)
    return L

def H(x, y, px, py):
    xsq, ysq = x**2, y**2
    LHt1 = 0.5*(px**2 + py**2)
    LHt2 = 0.5*(xsq + ysq)
    LHt3 = xsq*y - (y**3)/3
    LH = LHt1 + LHt2 + LHt3
    return(LH)

def hennon_hailes_loss(X , y, ydot, out_weights, force_t = fforce, 
                reg = True, ode_coefs = None, mean = True,
               enet_strength = None, enet_alpha = None, init_conds = None, lam = 1):
    

    #4 outputs
    assert init_conds

    x, y, px, py = [y[:,i].view(-1,1) for i in range(y.shape[1])]#, y[:,1].view(-1,1)
    dx, dy, dpx, dpy = [ydot[:,i].view(-1,1)for i in range(ydot.shape[1])]

    x0, y0, px0, py0 = init_conds

    xsq, ysq = x**2, y**2
    
    #with paramization
    term1 = (dx - px)**2 
    term2 = (dy - py)**2
    term3 = (dpx + x + 2*x*y)**2 
    term4 = (dpy + y + xsq - ysq )**2
    L = (term1 + term2 + term3 + term4).mean().pow(2)

    
    #hamiltonian regularization
    LH_ = H(x, y, px, py)
    LH0 = H(x0, y0, px0, py0)
    L = L + ((LH_ -LH0).mean()).pow(2)


    #if reg:
    weight_size_sq = torch.mean(torch.square(out_weights))
    weight_size_L1 = torch.mean(torch.abs(out_weights))
    L_reg = enet_strength*(enet_alpha * weight_size_sq + (1- enet_alpha) * weight_size_L1)
    L = L + 0.1 * L_reg 
    
    return L

def elastic_loss(X , y, ydot, out_weights, force_t = fforce, 
                reg = True, ode_coefs = None, mean = True,
               enet_strength = None, enet_alpha = None, init_conds = None, lam = 1):
    y, p = y[:,0].view(-1,1), y[:,1].view(-1,1)
    ydot, pdot = ydot[:,0].view(-1,1), ydot[:,1].view(-1,1)
    
    #with paramization
    L =  (ydot - p)**2 + (pdot + y + lam * y**3   - force_t)**2
    
    #if mean:
    L = torch.mean(L)
    #    hi = float(L)
    
    #if reg:
    weight_size_sq = torch.mean(torch.square(out_weights))
    weight_size_L1 = torch.mean(torch.abs(out_weights))
    L_reg = enet_strength*(enet_alpha * weight_size_sq + (1- enet_alpha) * weight_size_L1)
    L = L + 0.1 * L_reg 
    

#     y0, p0 = init_conds
#     ham = hamiltonian(y, p)
#     ham0 = hamiltonian(y0, p0)
#     L_H = (( ham - ham0).pow(2)).mean()
#     assert L_H >0

#     L = L +  L_H
    
    #print("L1", hi, "L_elastic", L_reg, "L_H", L_H)
    return L

def freparam(t, order = 1):
    exp_t = torch.exp(-t)
    
    derivatives_of_g = []
    
    g = 1 - exp_t
    
    #0th derivative
    derivatives_of_g.append(g)
    
    g_dot = 1 - g
    return g, g_dot

def driven_pop_loss(X , y, ydot, out_weights, lam = 1, force_t = None, reg = False, 
               ode_coefs = None, init_conds = None, 
                enet_alpha = None, enet_strength =None, mean = True):
    
    #with paramization
    L =  ydot  + lam * y - force_t
    
    # if reg:
    #     #assert False
    #     weight_size_sq = torch.mean(torch.square(out_weights))
    #     weight_size_L1 = torch.mean(torch.abs(out_weights))
    #     L_reg = enet_strength*(enet_alpha * weight_size_sq + (1- enet_alpha) * weight_size_L1)
    #     L = L + 0.1 * L_reg 
    
    L = torch.square(L)
    if mean:
        L = torch.mean(L)
    return L

def ham_loss(X , y, ydot, out_weights, force_t = None, 
                reg = True, ode_coefs = None, mean = True,
               enet_strength = None, enet_alpha = None, init_conds = None, lam = 1):
    y, p = y[:,0].view(-1,1), y[:,1].view(-1,1)
    ydot, pdot = ydot[:,0].view(-1,1), ydot[:,1].view(-1,1)
    
    #with paramization
    L =  (ydot - p)**2 + (pdot + y + lam * y**3   - force_t)**2
    
    #if mean:
    L = torch.mean(L)
    #    hi = float(L)
    
    #if reg:
    weight_size_sq = torch.mean(torch.square(out_weights))
    weight_size_L1 = torch.mean(torch.abs(out_weights))
    L_reg = enet_strength*(enet_alpha * weight_size_sq + (1- enet_alpha) * weight_size_L1)
    L = L + 0.1 * L_reg 
    

    y0, p0 = init_conds
    ham = hamiltonian(y, p)
    ham0 = hamiltonian(y0, p0)
    L_H = (( ham - ham0).pow(2)).mean()
    assert L_H >0

    L = L +  L_H
    
    #print("L1", hi, "L_elastic", L_reg, "L_H", L_H)
    return L

global lr_100_epoch
global lr_1000_epoch
global lr_5000_epoch

lr_100_epoch, lr_1000_epoch, lr_5000_epoch = [], [], []   

def optimize_last_layer(esn, 
                        SAVE_AFTER_EPOCHS = 1000,
                        epochs = 1000,#45000,
                        custom_loss = None,
                        loss_threshold = 10**-10,#10 ** -8,
                        EPOCHS_TO_TERMINATION = None,
                        f = fforce,
                        lr = 0.05, 
                        reg = None,
                        plott = False,
                        plot_every_n_epochs = 2499):#gamma 0.1, spikethreshold 0.07 works
    #define new_x
    new_X = esn.extended_states.detach()
    spikethreshold = esn.spikethreshold

    #force detach states_dot
    esn.states_dot = esn.states_dot.detach().requires_grad_(False)

    #define criterion
    criterion = torch.nn.MSELoss()

    #assert esn.LinOut.weight.requires_grad and esn.LinOut.bias.requires_grad
    #assert not new_X.requires_grad

    #define previous_loss (could be used to do a convergence stop)
    previous_loss = 0

    #define best score so that we can save the best weights
    best_score = 0

    #define the optimizer
    optimizer = optim.Adam(esn.parameters(), lr = lr)
    
    #optimizer = torch.optim.SGD(model.parameters(), lr=100)
    if esn.gamma_cyclic:
        cyclic_scheduler = torch.optim.lr_scheduler.CyclicLR(optimizer, 10**-6, 0.01,
                                          gamma = esn.gamma_cyclic,#0.9999,
                                          mode = "exp_range", cycle_momentum = False)
        
    scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=1, gamma=esn.gamma)
    lrs = []

    #define the loss history
    loss_history = []
    
    if plott:
        #use pl for live plotting
        fig, ax = pl.subplots(1,3, figsize = (16,4))
    
    t = esn.X#.view(*N.shape).detach()
    g, g_dot = esn.G
    y0  = esn.init_conds[0]
    
    flipped = False
    flipped2 = False
    pow_ = -4
    floss_last = 0
    
    
    try:
        assert esn.LinOut.weight.requires_grad and esn.LinOut.bias.requires_grad
    except:
        esn.LinOut.weight.requires_grad_(True)
        esn.LinOut.bias.requires_grad_(True)
        
    #bail
    
    #begin optimization loop
    for e in range(epochs):

        optimizer.zero_grad()
        
        N = esn.forward( esn.extended_states )
        N_dot = esn.calc_Ndot(esn.states_dot)

        y = g *N 
        
        ydot = g_dot * N + g * N_dot
        
        y[:,0] = y[:,0] + esn.init_conds[0]
        y[:,1] = y[:,1] + esn.init_conds[1]

        assert N.shape == N_dot.shape, f'{N.shape} != {N_dot.shape}'
        
        #assert esn.LinOut.weight.requires_grad and esn.LinOut.bias.requires_grad

        #total_ws = esn.LinOut.weight.shape[0] + 1
        #weight_size_sq = torch.mean(torch.square(esn.LinOut.weight))
        
        loss = custom_loss(esn.X, y, ydot, esn.LinOut.weight, reg = reg, ode_coefs = esn.ode_coefs,
                          init_conds = esn.init_conds, enet_alpha= esn.enet_alpha, enet_strength = esn.enet_strength)
        loss.backward()
        optimizer.step()
        if esn.gamma_cyclic and e > 100 and e < 4000:
            cyclic_scheduler.step()
        lrs.append(optimizer.param_groups[0]["lr"])
        
        
        floss = float(loss)
        loss_history.append(floss)
        
        
        """
        if floss < 10**pow_ :
            scheduler.step()
            lrs.append(optimizer.param_groups[0]["lr"])
            #for param_group in optimizer.param_groups:
            #    print('lr', param_group['lr'])
            pow_ -= 1
        """
#         if e == 10**3:
#             if floss > 10**(5):
#                 EPOCHS_TO_TERMINATION = e + 50
        
#         if e == 10**4:
#             if floss > 10**(2.5):
#                 EPOCHS_TO_TERMINATION = e + 50
        
        if e > 0:
            
            loss_delta = float(np.log(floss_last) - np.log(floss)) 
            if loss_delta > esn.spikethreshold:# or loss_delta < -3:
                lrs.append(optimizer.param_groups[0]["lr"])
                scheduler.step()
        
                
        quant = 0.8
        if e == 100:
            lr_100_epoch.append(floss)
            if floss >= np.quantile(np.array(lr_100_epoch), 0.8):
                if len(lr_100_epoch) > 20:
                    EPOCHS_TO_TERMINATION = e + 50
        elif e == 1000:
            lr_1000_epoch.append(floss)
            if floss >= np.quantile(np.array(lr_1000_epoch), 0.8):
                if len(lr_100_epoch) > 20:
                    EPOCHS_TO_TERMINATION = e + 50
        if e == 5000:
            lr_5000_epoch.append(floss)
            if floss >= np.quantile(np.array(lr_5000_epoch), 0.8):
                if len(lr_100_epoch) > 20:
                    EPOCHS_TO_TERMINATION = e + 50
        
        
        if not e and not best_score:
            best_bias, best_weight, best_fit = esn.LinOut.bias.detach(), esn.LinOut.weight.detach(), y.clone()

        if e > SAVE_AFTER_EPOCHS:
            if not best_score:
                
                if floss <= min(loss_history):
                    best_fit = y.clone()
                    best_ydot = ydot.clone()
                    best_bias, best_weight = esn.LinOut.bias.detach(), esn.LinOut.weight.detach()
                    best_score = float(loss)
            else:
                if floss < best_score:
                    best_fit = y.clone()
                    best_ydot = ydot.clone()
                    best_bias, best_weight = esn.LinOut.bias.detach(), esn.LinOut.weight.detach()
                    best_score = float(loss)
                    
        if not EPOCHS_TO_TERMINATION:
            if float(loss) < loss_threshold:
                EPOCHS_TO_TERMINATION = e + 100
        else:
            if e >= EPOCHS_TO_TERMINATION:
                return {"weights": best_weight, "bias" : best_bias, "fit" : best_fit, 
                        "loss" : loss_history, "best_score" : torch.tensor(best_score),
                        "RC" : esn}
        floss_last = floss
        if plott and e:

            if e % plot_every_n_epochs == 0:
                for param_group in optimizer.param_groups:
                    print('lr', param_group['lr'])
                ax[0].clear()
                logloss_str = 'Log(L) ' + '%.2E' % Decimal((loss).item())
                delta_loss  = ' delta Log(L) ' + '%.2E' % Decimal((loss-previous_loss).item())

                print(logloss_str + ", " + delta_loss)
                ax[0].plot(y.detach().cpu(), label = "exact")
                ax[0].set_title(f"Epoch {e}" + ", " + logloss_str)
                ax[0].set_xlabel("t")

                ax[1].set_title(delta_loss)
                ax[1].plot(N_dot.detach().cpu())
                #ax[0].plot(y_dot.detach(), label = "dy_dx")
                ax[2].clear()
                #weight_size = str(weight_size_sq.detach().item())
                #ax[2].set_title("loss history \n and "+ weight_size)

                ax[2].loglog(loss_history)
                ax[2].set_xlabel("t")

                [ax[i].legend() for i in range(3)]
                previous_loss = loss.item()

                #clear the plot outputt and then re-plot
                display.clear_output(wait=True) 
                display.display(pl.gcf())
            
            
    return {"weights": best_weight, "bias" : best_bias, "fit" : best_fit, 
            "loss" : loss_history, "best_score" : torch.tensor(best_score),
            "RC" : esn}




# -*- coding: utf-8 -*-
"""
Miscellaneous utilities and tools
"""
from __future__ import absolute_import, print_function

import functools
import keyword
import os
import re
import sys
from contextlib import contextmanager
from operator import itemgetter
from distutils.filelist import FileList

from six import PY2

from .templates import licenses
from .contrib import scm_setuptools_too_old


@contextmanager
def chdir(path):
    """Contextmanager to change into a directory

    Args:
        path (str): path to change current working directory to
    """
    curr_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(curr_dir)


def is_valid_identifier(string):
    """Check if string is a valid package name

    Args:
        string (str): package name

    Returns:
        bool: True if string is valid package name else False
    """
    if not re.match("[_A-Za-z][_a-zA-Z0-9]*$", string):
        return False
    if keyword.iskeyword(string):
        return False
    return True


def make_valid_identifier(string):
    """Try to make a valid package name identifier from a string

    Args:
        string (str): invalid package name

    Returns:
        str: valid package name as string or :obj:`RuntimeError`

    Raises:
        :obj:`RuntimeError`: raised if identifier can not be converted
    """
    string = string.strip()
    string = string.replace("-", "_")
    string = string.replace(" ", "_")
    string = re.sub('[^_a-zA-Z0-9]', '', string)
    string = string.lower()
    if is_valid_identifier(string):
        return string
    else:
        raise RuntimeError("String cannot be converted to a valid identifier.")


def list2str(lst, indent=0, brackets=True, quotes=True, sep=','):
    """Generate a Python syntax list string with an indention

    Args:
        lst ([str]): list of strings
        indent (int): indention
        brackets (bool): surround the list expression by brackets
        quotes (bool): surround each item with quotes
        sep (str): separator for each item

    Returns:
        str: string representation of the list
    """
    if quotes:
        lst_str = str(lst)
        if not brackets:
            lst_str = lst_str[1:-1]
    else:
        lst_str = ', '.join(lst)
        if brackets:
            lst_str = '[' + lst_str + ']'
    lb = '{}\n'.format(sep) + indent*' '
    return lst_str.replace(', ', lb)


def exceptions2exit(exception_list):
    """Decorator to convert given exceptions to exit messages

    This avoids displaying nasty stack traces to end-users

    Args:
        exception_list [Exception]: list of exceptions to convert
    """
    def exceptions2exit_decorator(func):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except tuple(exception_list) as e:
                print("ERROR: {}".format(e))
                sys.exit(1)
        return func_wrapper
    return exceptions2exit_decorator


# from http://en.wikibooks.org/, Creative Commons Attribution-ShareAlike 3.0
def levenshtein(s1, s2):
    """Calculate the Levenshtein distance between two strings

    Args:
        s1 (str): first string
        s2 (str): second string

    Returns:
        int: distance between s1 and s2
    """
    if len(s1) < len(s2):
        return levenshtein(s2, s1)

    # len(s1) >= len(s2)
    if len(s2) == 0:
        return len(s1)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def best_fit_license(txt):
    """Finds proper license name for the license defined in txt

    Args:
        txt (str): license name

    Returns:
        str: license name
    """
    ratings = {lic: levenshtein(txt, lic.lower()) for lic in licenses}
    return min(ratings.items(), key=itemgetter(1))[0]


def utf8_encode(string):
    """Encode a Python 2 unicode object to str for compatibility with Python 3

    Args:
        string (str): Python 2 unicode object or Python 3 str object

    Returns:
        str: Python 2 str object or Python 3 str object
    """
    return string.encode(encoding='utf8') if PY2 else string


def utf8_decode(string):
    """Decode a Python 2 str object to unicode for compatibility with Python 3

    Args:
        string (str): Python 2 str object or Python 3 str object

    Returns:
        str: Python 2 unicode object or Python 3 str object
    """
    return string.decode(encoding='utf8') if PY2 else string


def get_files(pattern):
    """Retrieve all files in the current directory by a pattern.

    Use ** as greedy wildcard and * as non-greedy wildcard.

    Args:
        pattern (str): pattern as used by :obj:`distutils.filelist.Filelist`

    Returns:
        [str]: list of files
    """
    filelist = FileList()
    if '**' in pattern:
        pattern = pattern.replace('**', '*')
        anchor = False
    else:
        anchor = True
    filelist.include_pattern(pattern, anchor)
    return filelist.files


def prepare_namespace(namespace_str):
    """Check the validity of namespace_str and split it up into a list

    Args:
        namespace_str (str): namespace, e.g. "com.blue_yonder"

    Returns:
        [str]: list of namespaces, e.g. ["com", "com.blue_yonder"]

    Raises:
          :obj:`RuntimeError` : raised if namespace is not valid
    """
    namespaces = namespace_str.split('.') if namespace_str else list()
    for namespace in namespaces:
        if not is_valid_identifier(namespace):
            raise RuntimeError(
                "{} is not a valid namespace package.".format(namespace))
    return ['.'.join(namespaces[:i+1]) for i in range(len(namespaces))]


def check_setuptools_version():
    """Check that setuptools has all necessary capabilities for setuptools_scm

    Raises:
          :obj:`RuntimeError` : raised if necessary capabilities are not met
    """
    if scm_setuptools_too_old:
        raise RuntimeError(
            "Your setuptools version is too old (<12). "
            "Use `pip install -U setuptools` to upgrade.\n"
            "If you have the deprecated `distribute` package installed "
            "remove it or update to version 0.7.3.")



# Based on the source code of GPyOpt/methods/modular_bayesian_optimization.py,
# which is licensed under the BSD 3-clause license

import time
import numpy as np
from GPyOpt.core.bo import BO
from GPyOpt.util.general import best_value
from GPyOpt.methods.modular_bayesian_optimization import ModularBayesianOptimization
from GPyOpt.plotting.plots_bo import plot_acquisition, plot_convergence

__all__ = ['EchoStateBO']


class EchoStateBO(BO):
    """
    Modular Bayesian optimization, edited for stepwise application in Echo State Networks. 
    This class wraps the optimization loop around the different handlers.

    Adapted from ModularBayesianOptimization (Copyright (c) 2016, the GPyOpt Authors, BSD 3 license)

    :param model: GPyOpt model class.
    :param space: GPyOpt space class.
    :param objective: GPyOpt objective class.
    :param acquisition: GPyOpt acquisition class.
    :param evaluator: GPyOpt evaluator class.
    :param X_init: 2d numpy array containing the initial inputs (one per row) of the model.
    :param Y_init: 2d numpy array containing the initial outputs (one per row) of the model.
    :param cost: GPyOpt cost class (default, none).
    :param normalize_Y: whether to normalize the outputs before performing any optimization (default, True).
    :param model_update_interval: interval of collected observations after which the model is updated (default, 1).
    """

    def __init__(self, model, space, objective, acquisition, evaluator, X_init, Y_init=None, model_update_interval=1):

        self.initial_iter = True
        self.modular_optimization = True

        # Create optimization space
        super(EchoStateBO, self).__init__(
            model=model,
            space=space,
            objective=objective,
            acquisition=acquisition,
            evaluator=evaluator,
            X_init=X_init,
            Y_init=Y_init,
            cost=None,
            normalize_Y=False,  # Normalization done by RobustGPModel
            model_update_interval=model_update_interval)

    def run_target_optimization(self, target_score=0., max_iter=2048, max_time=np.inf, eps=1e-8, verbosity=True):
        """
        Runs Bayesian Optimization for a number 'max_iter' of iterations (after the initial exploration data)
        :param max_iter: exploration horizon, or number of acquisitions.
            If nothing is provided optimizes the current acquisition.
        :param max_time: maximum exploration horizon in seconds.
        :param eps: minimum distance between two consecutive x's to keep running the model.
        :param verbosity: flag to print the optimization results after each iteration (default, True).
        :param report_file: filename of the file where the results of the optimization are saved (default, None).
        """
        # Save the options to print and save the results
        self.verbosity = verbosity

        # Avoid stacking: create full X and Y
        X_full = np.full((max_iter, self.X.shape[1]), fill_value=np.nan, dtype=np.float32)
        Y_full = np.full((max_iter, 1), fill_value=np.inf, dtype=np.float32)

        # Setting up stop conditions
        self.eps = eps
        self.max_iter = max_iter
        self.max_time = max_time
        self.target = target_score

        # Initialize iterations and running time
        self.time_zero = time.time()
        self.cum_time = 0
        self.num_acquisitions = 0

        # Initial function evaluation and model fitting
        if self.X is not None and self.Y is None:
            samples = self.X.shape[0]
            Y, _ = self.objective.evaluate(self.X)
            X_full[:samples, :] = self.X  # Confusingly, self.X and self.Y here are initial samples
            Y_full[:samples, :] = Y
            self.num_acquisitions += samples

        # Overwrite X and Y with full arrays to prevent stacking (slow)
        self.X = X_full
        self.Y = Y_full

        first_time = True

        # Initialize time cost of the evaluations
        while (self.max_time > self.cum_time):

            # Update GP model
            self._update_model()

            # Inform user
            if first_time:
                if self.verbosity:
                    print("Model initialization done.", '\n')
                    print(self.model.model, '\n')
                    print(self.model.model.kern.lengthscale, '\n')
                first_time = False

            # Update and optimize acquisition and compute the exploration level in the next evaluation
            self.suggested_sample = self._compute_next_evaluations()

            # Augment X
            i = self.num_acquisitions
            sample_size = self.suggested_sample.shape[0]
            space_left = self.X.shape[0] - i
            allocation = np.clip(sample_size, 0, space_left)
            self.X[i:i + allocation, :] = self.suggested_sample[:allocation]

            # Evaluate *f* in X, augment Y
            Y_new = self.evaluate_objective()
            self.Y[i:i + allocation, :] = Y_new[:allocation]

            # Update current evaluation time and function evaluations
            self.cum_time = time.time() - self.time_zero
            distance = self._distance_last_evaluations(sample_size)
            self.num_acquisitions += allocation

            # Target check
            if np.any(Y_new <= self.target):
                if self.verbosity:
                    pass
                print('Target reached at iteration', self.num_acquisitions)
                break

            # Check maximum iterations
            if self.num_acquisitions >= self.max_iter:
                if self.max_iter > 0 and self.verbosity:
                    print('Maximum iterations reached')
                break

            # Convergence
            if distance <= self.eps:
                if self.verbosity:
                    pass
                print('Converged at iteration', self.num_acquisitions)
                break

        # Stop messages and execution time
        self._compute_results()

        return self.num_acquisitions

    def evaluate_objective(self):
        """
        Evaluates the objective
        """
        Y_new, _ = self.objective.evaluate(self.suggested_sample)
        return Y_new

    def _compute_next_evaluations(self):
        """
        Computes the location of the new evaluation (optimizes the acquisition in the standard case).
        """
        return self.evaluator.compute_batch()

    def _distance_last_evaluations(self, sample_size):
        """
        Computes the distance between the last two evaluations.
        """
        i = self.num_acquisitions
        last = self.X[i, :]
        previous = self.X[i - sample_size, :]  # in Batch mode only the first sample is non-random
        distance = np.linalg.norm(last - previous, ord=1)  # Manhattan distance
        return distance

    def _update_model(self):
        """
        Updates the model (when more than one observation is available) and saves the parameters (if available).
        """
        i = self.num_acquisitions
        if i % self.model_update_interval == 0:
            self.model.updateModel(self.X[:i], self.Y[:i])

    def _compute_results(self):
        """
        Computes the optimum and its value.
        """
        i = self.num_acquisitions
        Y = self.Y[:i]
        self.Y_best = best_value(Y)
        self.x_opt = self.X[np.argmin(Y), :]
        self.fx_opt = np.min(Y)

    def plot_convergence(self, filename=None):
        """
        Makes twp plots to evaluate the convergence of the model:
            plot 1: Iterations vs. distance between consecutive selected x's
            plot 2: Iterations vs. the mean of the current model in the selected sample.
        :param filename: name of the file where the plot is saved
        """
        i = self.Y_best.shape[0]
        return plot_convergence(self.X[:i], self.Y_best, filename)

    def get_evaluations(self):
        i = self.num_acquisitions
        return self.X[:i], self.Y[:i]



from setuptools_scm import format_version
from setuptools_scm.hg import archival_to_version, parse
from setuptools_scm import integration

import pytest


@pytest.fixture
def wd(wd):
    wd('hg init')
    wd.add_command = 'hg add .'
    wd.commit_command = 'hg commit -m test-{reason} -u test -d "0 0"'
    return wd


archival_mapping = {
    '1.0': {'tag': '1.0'},
    '1.1.dev3+h000000000000': {
        'latesttag': '1.0',
        'latesttagdistance': '3',
        'node': '0'*20,
    },
    '0.0': {
        'node': '0'*20,
    },
    '1.2.2': {'tag': 'release-1.2.2'},
    '1.2.2.dev0': {'tag': 'release-1.2.2.dev'},

}


@pytest.mark.parametrize('expected,data', sorted(archival_mapping.items()))
def test_archival_to_version(expected, data):
    version = archival_to_version(data)
    assert format_version(
        version,
        version_scheme='guess-next-dev',
        local_scheme='node-and-date') == expected


def test_find_files_stop_at_root_hg(wd):
    wd.commit_testfile()
    wd.cwd.ensure('project/setup.cfg')
    assert integration.find_files(str(wd.cwd/'project')) == []


# XXX: better tests for tag prefixes
def test_version_from_hg_id(wd):
    assert wd.version == '0.0'

    wd.commit_testfile()
    assert wd.version.startswith('0.1.dev2+')

    # tagging commit is considered the tag
    wd('hg tag v0.1 -u test -d "0 0"')
    assert wd.version == '0.1'

    wd.commit_testfile()
    assert wd.version.startswith('0.2.dev2')

    wd('hg up v0.1')
    assert wd.version == '0.1'

    # commit originating from the taged revision
    # that is not a actual tag
    wd.commit_testfile()
    assert wd.version.startswith('0.2.dev1+')

    # several tags
    wd('hg up')
    wd('hg tag v0.2 -u test -d "0 0"')
    wd('hg tag v0.3 -u test -d "0 0" -r v0.2')
    assert wd.version == '0.3'


def test_version_from_archival(wd):
    # entrypoints are unordered,
    # cleaning the wd ensure this test wont break randomly
    wd.cwd.join('.hg').remove()
    wd.write(
        '.hg_archival.txt',
        'node: 000000000000\n'
        'tag: 0.1\n'
    )
    assert wd.version == '0.1'

    wd.write(
        '.hg_archival.txt',
        'node: 000000000000\n'
        'latesttag: 0.1\n'
        'latesttagdistance: 3\n'
    )

    assert wd.version == '0.2.dev3+h000000000000'


@pytest.mark.issue('#72')
def test_version_in_merge(wd):
    wd.commit_testfile()
    wd.commit_testfile()
    wd('hg up 0')
    wd.commit_testfile()
    wd('hg merge --tool :merge')
    assert wd.version is not None


@pytest.mark.issue(128)
def test_parse_no_worktree(tmpdir):
    ret = parse(str(tmpdir))
    assert ret is None


@pytest.fixture
def version_1_0(wd):
    wd('hg branch default')
    wd.commit_testfile()
    wd('hg tag 1.0 -u test -d "0 0"')
    return wd


@pytest.fixture
def pre_merge_commit_after_tag(wd, version_1_0):
    wd('hg branch testbranch')
    wd.write('branchfile', 'branchtext')
    wd(wd.add_command)
    wd.commit()
    wd('hg update default')
    wd('hg merge testbranch')
    return wd


@pytest.mark.usefixtures("pre_merge_commit_after_tag")
def test_version_bump_before_merge_commit(wd):
    assert wd.version.startswith('1.1.dev1+')


@pytest.mark.issue(219)
@pytest.mark.usefixtures("pre_merge_commit_after_tag")
def test_version_bump_from_merge_commit(wd):
    wd.commit()
    assert wd.version.startswith('1.1.dev3+')  # issue 219


@pytest.mark.usefixtures("version_1_0")
def test_version_bump_from_commit_including_hgtag_mods(wd):
    """ Test the case where a commit includes changes to .hgtags and other files
    """
    with wd.cwd.join('.hgtags').open('a') as tagfile:
        tagfile.write('0  0\n')
    wd.write('branchfile', 'branchtext')
    wd(wd.add_command)
    assert wd.version.startswith('1.1.dev1+')  # bump from dirty version
    wd.commit()  # commits both the testfile _and_ .hgtags
    assert wd.version.startswith('1.1.dev2+')



from collections import OrderedDict
import copy
from dataclasses import dataclass
import json
from math import ceil, fabs
import numpy as np
from scipy.sparse import csr_matrix
from multiprocessing import Pool as mp_Pool
import multiprocessing
import pylab as pl
from IPython import display

from .esn import *
from copy import deepcopy

from scipy.sparse import csc_matrix
from scipy.sparse import csr_matrix

import types
import functools

from .defs import *

import logging
#logger = multiprocessing.log_to_stderr()
#logger.setLevel(multiprocessing.SUBDEBUG)



#https://stackoverflow.com/questions/9336646/python-decorator-with-multiprocessing-fails
class my_decorator(object):
    def __init__(self, target):
        self.target = target
        try:
            functools.update_wrapper(self, target)
        except:
            pass

    def __call__(self, candidates, args):
        f = []
        for candidate in candidates:
            f.append(self.target([candidate], args)[0])
        return f

# ideas from yesterday (trip related)
# 1. moving average RC
# 2. Penalize the time the algorithm takes to run

@dataclass
class TurboState:
    """
    This is from BOTorch. The Turbo state is a stopping condition.

    #TODO finish description and read Turbo paper

    Arguments:
        dim: 
        batch_size:
        length_min:
        length_max:
        failure_counter:
        success_counter:
        success_tolerance:
        best_value:        the best value we have seen so far
        restart_triggered: has a restart been triggered? If yes BO will terminate.
    """
    dim: int
    batch_size: int
    length: float = 0.8
    length_min: float = 0.5 ** 7
    length_max: float = 1.6
    failure_counter: int = 0
    failure_tolerance: int = float("nan")  # Note: Post-initialized
    success_counter: int = 0
    success_tolerance: int = 10 # Note: The original paper uses 3
    best_value: float = -float("inf")
    restart_triggered: bool = False

    def __post_init__(self):
        self.failure_tolerance = ceil(
            max([4.0 / self.batch_size, float(self.dim) / self.batch_size])
        )

def get_initial_points(dim, n_pts, device, dtype):
    sobol = SobolEngine(dimension=dim, scramble=True)
    X_init = sobol.draw(n=n_pts).to(dtype=dtype, device=device)
    return X_init

def update_state(state, Y_next):
    """Updates the turbo state (checks for stopping condition)
    Essentially this checks our TURBO stopping condition.
    
    Arguments:
        state:  the Turbo state
        Y_next: the most recent error return by the objective function

    """
    if max(Y_next) > state.best_value + 1e-3 * fabs(state.best_value):
        state.success_counter += 1
        state.failure_counter = 0
    else:
        state.success_counter = 0
        state.failure_counter += 1

    if state.success_counter == state.success_tolerance:  # Expand trust region
        state.length = min(2.0 * state.length, state.length_max)
        state.success_counter = 0
    elif state.failure_counter == state.failure_tolerance:  # Shrink trust region
        state.length /= 2.0
        state.failure_counter = 0

    state.best_value = max(state.best_value, max(Y_next).item())
    if state.length < state.length_min:
        state.restart_triggered = True

    return state


def generate_batch(
    state,
    model,
    X,  
    Y,
    batch_size,
    n_candidates=None,
    num_restarts=10,
    raw_samples=512,
    acqf="ts",  # "ei" or "ts",
    dtype = torch.float32,
    device = None
):
    """generate a batch for the Bayesian Optimization

    Arguments:
        state: the TURBO state (a stopping metric)
        model: the GP (Gaussian Process) BOTorch model
        X: points evaluated (a vector of hyper-parameter values fed to the objective function) # Evaluated points on the domain [0, 1]^d in original example, not ours.
        Y: Function values
        n_candidates: Number of candidates for Thompson sampling
        num_restarts:
        raw_samples:
        acqf: acquisition function (thompson sampling is preferred)
    """
    assert acqf in ("ts", "ei")
    assert X.min() >= 0.0 and X.max() <= 1.0 and torch.all(torch.isfinite(Y))
    if n_candidates is None:
        n_candidates = min(5000, max(2000, 200 * X.shape[-1]))

    # Scale the TR to be proportional to the lengthscales
    x_center = X[Y.argmax(), :].clone()
    weights = model.covar_module.base_kernel.lengthscale.squeeze().detach()
    weights = weights / weights.mean()
    weights = weights / torch.prod(weights.pow(1.0 / len(weights)))
    tr_lb = clamp(x_center - weights * state.length / 2.0, 0.0, 1.0)
    tr_ub = clamp(x_center + weights * state.length / 2.0, 0.0, 1.0)

    if acqf == "ts":
        dim = X.shape[-1]
        sobol = SobolEngine(dim, scramble=True)
        pert = sobol.draw(n_candidates, dtype=dtype).to(dtype=dtype, device=device)
        pert = tr_lb + (tr_ub - tr_lb) * pert

        # Create a perturbation mask
        prob_perturb = min(20.0 / dim, 1.0)
        mask = (
            rand(n_candidates, dim, dtype=dtype, device=device)
            <= prob_perturb
        )
        ind = torch.where(mask.sum(dim=1) == 0)[0]
        mask[ind, randint(0, dim - 1, size=(len(ind),), device=device)] = 1

        # Create candidate points from the perturbations and the mask        
        X_cand = x_center.expand(n_candidates, dim).clone()
        X_cand[mask] = pert[mask]

        # Sample on the candidate points
        thompson_sampling = MaxPosteriorSampling(model=model, replacement=False)
        X_next = thompson_sampling(X_cand, num_samples=batch_size)

    elif acqf == "ei":
        ei = qExpectedImprovement(model, train_Y.max(), maximize=True)
        X_next, acq_value = optimize_acqf(
            ei,
            bounds=stack([tr_lb, tr_ub]),
            q=batch_size,
            num_restarts=num_restarts,
            raw_samples=raw_samples,
        )

    return X_next

class SparseBooklet:
    """A set of preloaded reservoir weights matching the reservoir seed and approximate sparcity (if applicable)

    Parameters
    ----------
    book: dict
        the set of approximate reservoir tensors and associated connectivity thresholds
    keys:
        the key to access the dictionary (typically an approximate sparcity threshold)
    """
    def __init__(self, book, keys):
        self.sparse_book = book
        self.sparse_keys_ = np.array(keys)

    def get_approx_preRes(self, connectivity_threshold):
        """ Given a connectivity threshold, the method will return the sparse matrix most closely matching that threshold.

        Parameters
        ----------
        connectivity_threshold: float
            #TODO description

        """
        #print("sparse_keys", self.sparse_keys_, "connectivity_threshold", connectivity_threshold   )
        key_ =  self.sparse_keys_[self.sparse_keys_ > connectivity_threshold][0]
        val =  self.sparse_book[key_].clone()
        return val

class GlobalSparseLibrary:
    """
    This will approximate the search for the sparcity hyper-parameter, which will dramatically speed up training of the network.

    Parameters
    ----------
    lb: int
        lower bound (connectivity)
    ub: int
        upper bound (connectivity)
    n_nodes: number of nodes in the reservoir
    precision: the precision of the approximate sparcity metric
    flip_the_script: bool
        completely randomizes which reservoir has been selected.
    """

    def __init__(self, device, lb = -5, ub = 0, n_nodes = 1000, precision = None, 
                 flip_the_script = False):
        self.lb = lb
        self.ub = ub
        self.n_nodes_ = n_nodes
        self.library = {}
        self.book_indices = []
        self.precision = precision
        self.flip_the_script = flip_the_script
        self.device = device
        

    def addBook(self, random_seed):
        """
        Add a sparse reservoir set by looping through different different connectivity values and assigining one reservoir weight matrix per connetivity level
        and storing these for downstream use by EchoStateNetwork
        We generate the reservoir weights and store them in the sparse library.

        Parameters
        ----------
        random_seed: the random seed of the SparseLibrary with which to make the preloaded reservoir matrices
        """
        book = {}
        n = self.n_nodes_
        
        random_state = Generator(device = self.device).manual_seed(random_seed)

        accept = rand(n, n, generator = random_state, device = self.device) 
        reservoir_pre_weights = rand(n, n, generator = random_state, device = self.device) * 2 -1

        "for now we're going to avoid sparse matrices"
        for connectivity in np.logspace(self.ub, self.lb, self.precision): #, device = self.device):
            #book[connectivity] = csc_matrix((accept < connectivity ) * reservoir_pre_weights)
            
            book[connectivity] = (accept < connectivity ) * reservoir_pre_weights
        sparse_keys_ = sorted(book)

        self.library[random_seed] = SparseBooklet(book = book, keys = sparse_keys_)
        self.book_indices.append(random_seed)

    def getIndices(self):
        """returns book indices"""
        return self.book_indices

    def get_approx_preRes(self, connectivity_threshold, index = 0):
        """ This function is for use by EchoStateNetwork to access different sets of reservoir matrices.
        Given a connectivity threshold we access a reservoir by approximate sparcity / connectivity.
        But which randomly generated reservoir we select is determined by the index, which is what ESN uses if the one reservoir is nilpotent.
        Parameters
        ----------
            connectivity threshold: float
            index: int
                which preloaded reservoir do we want to load? Each index references a difference Sparse Booklet (ie a different reservoir)

        Returns
        -------
        sparse booklet for reading downstream by EchoStateNetwork class
        (we are returning a set of pre-loaded matrices to speed up optimization of the echo-state network by avoiding 
        repeated tensor generation.)
        """
        if self.flip_the_script:
            index = np.random.randint(len(self.book_indices))
        #print("index", index, "book indices", self.book_indices, "self.library", self.library)
        book = self.library[self.book_indices[index]]
        if index != 0:
            printc("retrieving book from library" + str(self.book_indices[index]), 'green')
        return book.get_approx_preRes(connectivity_threshold)


class ReservoirBuildingBlocks:
    """ An object that allows us to save reservoir components (independent of hyper-parameters) for faster optimization.


    Parameters
    ----------
        model_type: either random, cyclic or delay line
        input_weight_type: exponential or uniform
        random_seed: the random seed to set the reservoir
        n_nodes: the nodes of the network
        n_inputs: the number of observers in the case of a block experiment, the size of the output in the case of a pure prediction where teacher forcing is used.
    

    """
    def __init__(self, model_type, input_weight_type, random_seed , n_nodes, n_inputs = None, Distance_matrix = None, sparse = False, device = None):
        #print("INITIALING RESERVOIR")

        #initialize attributes
        self.device = device
        self.sparse = sparse
        self.random_seed = random_seed
        self.tensorArgs = {"device" : self.device}
        
        self.input_weight_type_ = input_weight_type
        self.model_type_ = model_type
        self.n_inputs_ = n_inputs
        n = self.n_nodes_ = n_nodes
        self.state = zeros(1, self.n_nodes_, device = self.device)

        if self.sparse:
            if model_type == "random":
                self.gen_ran_res_params()
                self.gen_sparse_accept_dict()
                assert 1 ==0
        else:
            random_state = Generator(device = self.device).manual_seed(random_seed)
            self.accept = rand(n, n, generator = random_state, device = self.device) 
            self.reservoir_pre_weights = rand(n, n, generator = random_state, device = self.device) * 2 -1


    def gen_ran_res_params(self):
        """Generates the matrices required for generating reservoir weights"""
        gen = Generator(device = self.device).manual_seed(self.random_seed)
        n = self.n_nodes_
        self.accept = rand(n, n, **self.tensorArgs, generator = gen)
        self.reservoir_pre_weights = rand(n, n, **self.tensorArgs, generator = gen) * 2 - 1

    def gen_sparse_accept_dict(self, reservoir_seeds = [123, 999], precision = 1000):
        """
        #TODO description

        Parameters
        ----------
        reservoir_seeds:
            preloaded reservoirs to generate (random seeds will uniquely create different reservoirs)
            precision: how precisely do you want to approximate connectivity in log space? 
        """
        printc("GENERATING SPARSE DICT", 'cyan')
        global sparse_dict
        
        #printc("Building approximate sparse reservoirs for faster optimization ...",'fail')
        #for connectivity in np.logspace(0, -5, precision):
        #    sparse_dict[connectivity] = csc_matrix((self.accept < connectivity ) * self.reservoir_pre_weights)
        #self.sparse_keys_ = np.array(sorted(sparse_dict))
        self.number_of_preloaded_sparse_sets = len(reservoir_seeds)
        sparse_dict = GlobalSparseLibrary(self.device, precision = precision, n_nodes = self.n_nodes_)
        for random_seed in reservoir_seeds:
            printc("generated sparse reservoir library for random seed " + str(random_seed), 'cyan')
            sparse_dict.addBook(random_seed)

    def get_approx_preRes(self, connectivity_threshold, i):
        """
        You can use the matrix returned instead of...
        """
        val = sparse_dict.get_approx_preRes(connectivity_threshold, index = i)
        return val

    def get_approx_preRes_old(self, connectivity_threshold, i):
        """
        You can use the matrix returned instead of...

        Parameters
        ----------
        inputs : dict
            contains start_index and random_seed. 
                    (+) start index determines where to start the cross-validated sample
                    (+) random seed defines the reservoir's random seed.
        Returns
        ----------
        dict of new training and validation sets
        """
        key_ =  self.sparse_keys_[self.sparse_keys_ > connectivity_threshold][0]
        val =  np.array(sparse_dict[key_]).copy()
        return val

    def gen_in_weights(self):
        """ Generates the reservoir input weight matrix
        
        This method assigns the reservoir input weights for later use downstream by the EchoStateNetworkClass.

        Parameters
        ----------
        inputs : None
        Returns: None
        ----------
        
        """

        gen = Generator(device = self.device).manual_seed(self.random_seed)
        n, m = self.n_nodes_, self.n_inputs_
        in_w_shape_ = (n, m)
        print('m,n', m,n)

        #at the moment all input weight matrices use uniform bias.
        self.bias = rand( n, 1, generator = gen, device = self.device) * 2 - 1

        #weights
        if self.input_weight_type_ == "uniform":
            self.in_weights = rand((n,m), generator = gen, device = self.device)
            self.in_weights = self.in_weights * 2 - 1
            print('in_weights', self.in_weights.shape)

        elif self.input_weight_type_ == "exponential":
            printc("BUILDING SIGN_", 'fail')
            sign1 = random_state.choice([-1, 1], size= (in_w_shape_[0], in_w_shape_[1]//2))
            sign2 = random_state.choice([-1, 1], size= (in_w_shape_[0], in_w_shape_[1]//2))

            self.sign_dual = (sign1, sign2)
            self.sign = np.concatenate((sign1, sign2), axis = 1)

        #regularization
        self.feedback_weights = rand(n, 1, **self.tensorArgs, generator = gen) * 2 - 1

        #regularization
        self.noise_z = normal(0, 1, size = (n, m), **self.tensorArgs, generator = gen)


__all__ = ['EchoStateNetworkCV']

def process_score(score__, upper_error_limit = 1000000, device = None):
                    if torch.isnan(score__):
                        score__ = tensor(upper_error_limit, device = device, requires_grad  = False, dtype = torch.float32)
                    else:
                        score__ = min(score__, tensor(upper_error_limit, device = device, requires_grad = False, dtype = torch.float32))
                    return score__

def combine_score(tr_score, val_score, tr_score_prop, log_score):
        tr_score = tr_score.type(torch.float32)
        val_score = val_score.type(torch.float32)
        if log_score:
            tr_score = torch.log(tr_score)
            val_score = torch.log(val_score)
            return tr_score * tr_score_prop + val_score * (1- tr_score_prop)
        else:
            return torch.log(tr_score * tr_score_prop + val_score * (1- tr_score_prop))

def execute_objective(arguments):
    """ Function at the heart of RCTorch, train a network on multiple rounds of cross-validated train/test info, then return the average error.
    This method also deals with dispatching mutliple series to the objective function if there are multiple, and aggregates the returned scores
        by averaging.

    Parameters
    ----------
        arguments: a list of arguments that have been put into dictionary form for multiprocessing convenience
        upper_error_limit: brutal error clipping upper bound, 'nan' error function returns the maximum upper limit as well to discourage
                the algorithm from searching in that part of the parameter space.

    Returns
    -------
        tuple of score (float), the prediction and validation sets for plotting (optional), and the job id

    (we need the job id to resort and relate X and y for BO Opt which have been scrambled by multiprocessing)
    """
    #later call define_tr_val from within this function for speedup.
    #score, pred_ = self.define_tr_val() #just get the train and test...
    

    cv_samples, parallel_arguments, parameters, windowsOS, id_ = arguments
    device = parallel_arguments["device"]
    declaration_args = parallel_arguments["declaration_args"]
    backprop_args = parallel_arguments["backprop_args"]
    #del declaration_args['reservoir']

    cv_args = parallel_arguments["cv_args"]
    log_score, rounds, tr_score_prop = cv_args["log_score"], cv_args["rounds"], cv_args["tr_score_prop"]

    # if windowsOS == True:
    #     #move specific arguments to the gpu.
    #     cv_samples_ = []
    #     for i, cv_sample in enumerate(cv_samples):
    #         if  cv_sample["tr_y"].device != device:
    #             #if cv_sample["tr_x"]:
    #             train_x, validate_x = cv_sample["tr_x"].to(device), cv_sample["val_x"].to(device)

    #             train_y, validate_y  = cv_sample["tr_y"].to(device), cv_sample["val_y"].to(device)
    #         else:
    #             #consider not sending the cv_sample["x"] if it's a pure prediction.
    #             train_x, train_y = cv_sample["tr_x"], cv_sample["tr_y"]
    #             validate_x, validate_y = cv_sample["val_x"], cv_sample["val_y"]
    #         cv_samples_.append((train_x, train_y, validate_x, validate_y))

    #         #now move the input weights and the reservoir arguments to the gpu.
    #         #deepcopy()
    #         assert declaration_args["reservoir"] != None
    #         if declaration_args["reservoir"] != None:
    #             declaration_args["reservoir"].in_weights = declaration_args["reservoir"].in_weights.to(device)
    #             declaration_args["reservoir"].accept = declaration_args["reservoir"].accept.to(device)
    #             declaration_args["reservoir"].reservoir_pre_weights = declaration_args["reservoir"].reservoir_pre_weights.to(device)

    #     RC = EchoStateNetwork(**declaration_args,  **parameters, id_ = id_)
    
    RC = EchoStateNetwork(**declaration_args, **parameters, id_ = id_)
    cv_samples_ = []
    for i, cv_sample in enumerate(cv_samples):
        cv_samples_.append((cv_sample["tr_x"],  cv_sample["tr_y"], cv_sample["val_x"], cv_sample["val_y"]))

    ODE_order = parallel_arguments["train_args"]["ODE_order"]

    train_args = parallel_arguments["train_args"]
    test_args = parallel_arguments["test_args"]

    total_score = 0

    backprop_f = backprop_args["backprop_f"]

    if type(train_args["init_conditions"][0]) == list:
        multiple_ICs = True if len(train_args["init_conditions"][0]) > 1 else False
    else:
        multiple_ICs = False

    for i, cv_sample in enumerate(cv_samples_):
        #print(i)
        train_x, train_y, validate_x, validate_y = cv_sample

        cv_sample_score = 0

        #divis = rounds *2
        if ODE_order:
            if multiple_ICs:
                # if train_score:
                #         return {"scores" : scores, 
                #                 "weights": gd_weights, 
                #                 "biases" : gd_biases,
                #                 "ys"     : ys,
                #                 "ydots"  : ydots,
                #                 "losses" : Ls}
                dictt = RC.fit(X = train_x, y = train_y, train_score = True, **train_args)
                train_scores = dictt["scores"]
                val_scores, pred_, id_ = RC.test(X=validate_x, y= validate_y, **test_args)

                
                for i, train_score in enumerate(train_scores):
                    
                    train_scores[i] = process_score(train_score, device = device)

                    val_score = process_score(val_scores[i], device = device)# / divis

                    round_score = combine_score(train_score, val_score, tr_score_prop, log_score)  
                    cv_sample_score += round_score 
                total_score += cv_sample_score

            else:
                #train_score = RC.fit(X = train_x, y = train_y, train_score = True, **parallel_arguments["train_args"])
                results = RC.fit(X = train_x, y = train_y, train_score = True, **train_args)
                train_scores = results["scores"]
                train_score = train_scores[0]

                train_score = process_score(train_score, device = device)# / divis

                val_scores, pred_, id_ = RC.test(X=validate_x, y= validate_y, **test_args)
                val_score = process_score(val_scores[0], device = device)# / divis

                cv_sample_score = combine_score(train_score, val_score, tr_score_prop, log_score)  

            
        else:
            _ = RC.fit(X = train_x, y = train_y,**parallel_arguments["train_args"])
            #train_score = process_score(train_score)
            
            val_score, pred_, id_ = RC.test(X=validate_x, y= validate_y, **parallel_arguments["test_args"])
            val_score = process_score(val_score)

            cv_sample_score = val_score #combine_score(train_score, val_score)

        #del train_x; del train_y;
        #del validate_x;

        if id_ != 0:
            del validate_y; del pred_;

        total_score += cv_sample_score
        #if device == torch.device('cuda'):
        #    torch.cuda.empty_cache()

    #del RC;
    #score_mu = total_score/len(cv_samples)
    #del cv_samples;
    nn = len(cv_samples_) * len(train_scores)

    total_score = total_score / nn
    if id_ == 0:
        try:
            validate_y = validate_y.to("cpu")
        except:
            pass
        return float(total_score), {"pred": pred_.to("cpu"), "val_y" : validate_y}, id_
    else:
        return float(total_score), None, id_

def sech2(z):
    return (1/(torch.cosh(z)))**2

class EchoStateNetworkCV:
    """A cross-validation object that automatically optimizes ESN hyperparameters using Bayesian optimization with
    Gaussian Process priors.

    Searches optimal solution within the provided bounds.

    The acquisition function currently implimented is Thompson Sampling.

    Parameters
    ----------
    bounds : dict
        A dictionary specifying the bounds for optimization. The key is the parameter name and the value
        is a tuple with minimum value and maximum value of that parameter. E.g. {'n_nodes': (100, 200), ...}
    model : class: {EchoStateNetwork}
            Model class to optimize
    subsequence_length : int
        Number of samples in one cross-validation sample
    initial_samples : int
        The number of random samples to explore the  before starting optimization
    validate_fraction : float
        The fraction of the data that may be used as a validation set
    batch_size : int
        Batch size of samples used by GPyOpt
    cv_samples : int
        Number of samples of the objective function to evaluate for a given parametrization of the ESN
    scoring_method : {'mse', 'rmse', 'tanh', 'nmse', 'nrmse', 'log', 'log-tanh', 'tanh-nrmse'}
        Evaluation metric that is used to guide optimization
    esn_burn_in : int
        Number of time steps to discard upon training a single Echo State Network
    esn_feedback : bool or None
        Build ESNs with feedback ('teacher forcing') if available
    verbose : bool
        Verbosity on or off

    device : string or torch device                                 #TODO flexible implimentation
        Torch device (either 'cpu' or 'cuda')
    interactive : bool
        if true, make interactive python plots. Useful in a jupyter notebook.
    approximate reservoir: bool
        if true, builds approximate sparse reservoirs and (ie approximate connectivity not precise). 
        It likely slightly reduces the final result's score but greatly speeds up the algorithm. #SPARCITY NOT IMPLIMENTED IN RCTORCH
    input_weight_type : string
        {"uniform"} is currently implimented. 
        #TODO: exponential and normal weights.
    activation function: nn.function
        The activation function used in the reservoir
    model_type: string
            #TODO
        right now it is unclear whether this means reservoir type or model type.
        likely that is unclear because I haven't implimented cyclic or exponential here. #TODO impliment uniform and expo weights
    BOTorch Turbo bayesian optimization parameters
    ----------------------------------------------
    failure tolerance: int
        the number of times that the model can fail to improve before length is in increased in turbo algo.
    success_tolerance: int
        like the explanation above this needs work.
    length_min: int
        The stopping condition. If the turbo_state's length falls below length_min then the algorithm will terminate.
    learning_rate: float
        if backprop is True, then the RC will train with gradient descent. In this case this is that learning rate.
    success_tolerance:
        #TODO description
    failure_tolerance:
        #TOD description

    #################### NOT IMPLIMENTED YET IN TORCH version (Hayden Fork)
    obs_index = None, target_index = None,  Distance_matrix = None, n_res = 1, 
    self.obs_index = obs_index
     self.target_index = target_index

    #################### NOT IMPLIMENTED IN TORCH version (came from Reinier)

    #################### eps, aquisition type and njobs seem like good things to port over.
    steps_ahead : int or None
        Number of steps to use in n-step ahead prediction for cross validation. `None` indicates prediction
        of all values in the validation array.
    max_iterations : int
        Maximim number of iterations in optimization
    log_space : bool
        Optimize in log space or not (take the logarithm of the objective or not before modeling it in the GP)
        ####### NOT IMPLIMENTED IN TORCH
    tanh_alpha : float
        Alpha coefficient used to scale the tanh error function: alpha * tanh{(1 / alpha) * mse}
    max_time : float
        Maximum number of seconds before quitting optimization
    acquisition_type : {'MPI', 'EI', 'LCB'}
        The type of acquisition function to use in Bayesian Optimization
    eps : float
        The number specifying the maximum amount of change in parameters before considering convergence
    plot : bool
        Show convergence plot at end of optimization
    target_score : float
        Quit when reaching this target score
    n_jobs : int
        Maximum number of concurrent jobs
    ####################

    #### uncompleted tasks:
    #1) upgrade to multiple acquisition functions.
    #self.acquisition_type = acquisition_type
    ######
    """
    def __init__(self, bounds, subsequence_length, model=EchoStateNetwork, initial_samples=50,
                 validate_fraction=0.5, steps_ahead=None, batch_size=1, cv_samples=1,
                 scoring_method='nrmse', esn_burn_in=0, random_seed=None, esn_feedback=None, 
                 verbose=True, model_type = "random", activation_function = nn.Tanh(), 
                 input_weight_type = "uniform", interactive = False, 
                 approximate_reservoir = False, length_min = 2**(-9), 
                 device = None, success_tolerance = 7, dtype = torch.float32,
                 windowsOS = False, track_in_grad = False, patience = 400, ODE_order = None,
                 dt = None, log_score =  False, act_f_prime = sech2
                 ):
        self.dt = dt
        self.log_score = log_score

        self.ODE_order = ODE_order
        self.windowsOS = windowsOS
        if not self.windowsOS:
            try:
                multiprocessing.set_start_method('spawn')
            except:
                pass
        if not device:
            self.device = torch_device("cuda" if cuda_is_available() else "cpu")
        else:
            self.device = device
        if self.device == torch_device('cuda'):
            torch.cuda.empty_cache()
        self.dtype = dtype

        self.track_in_grad = track_in_grad
        self.patience = patience

        print("FEEDBACK:", esn_feedback, ", device:", device)

        #turbo variables
        self.batch_size = batch_size
        self.length_min = length_min
        self.success_tolerance = success_tolerance

        #interactive plotting for jupyter notebooks.
        self.interactive = interactive
        self.approximate_reservoir = approximate_reservoir

        # Bookkeeping
        self.bounds = bounds
        self.parameters = OrderedDict(bounds) 
        self.errorz, self.errorz_step = [], []
        self.free_parameters = []
        self.fixed_parameters = []
        #self.n_res = n_res

        # Store settings
        self.model = model
        self.subsequence_length = subsequence_length
        self.initial_samples = initial_samples
        self.validate_fraction = validate_fraction
        self.steps_ahead = steps_ahead
        self.batch_size = batch_size
        self.cv_samples = cv_samples 
        self.esn_burn_in =  tensor(esn_burn_in, dtype=torch.int32).item()   #torch.adjustment required
        self.esn_feedback = esn_feedback

        self.seed = random_seed
        self.feedback = esn_feedback
        self.verbose = verbose
        self.model_type = model_type

        #self.Distance_matrix = Distance_matrix

        # Normalize bounds domains and remember transformation
        self.scaled_bounds, self.bound_scalings, self.bound_intercepts = self.normalize_bounds(self.parameters)
        
        self.activation_function = activation_function
        self.act_f_prime = act_f_prime

        self.input_weight_type = input_weight_type
        
        if "n_nodes" in self.bounds:
            if type(self.bounds["n_nodes"]) != int and type(self.bounds["n_nodes"]) != float:
                self.reservoir_matrices = None
                assert False, type(self.bounds["n_nodes"])
            else:
                self.reservoir_matrices = ReservoirBuildingBlocks(model_type = self.model_type, 
                                                              random_seed = self.seed,
                                                              n_nodes = self.bounds["n_nodes"],
                                                              input_weight_type = self.input_weight_type,
                                                              device = self.device)
        else:
            assert 1 == 0, "You must enter n_nodes as an argument into bounds_dict. ie: '\{ n_nodes: 1000 \}'"
            
        self.iteration_durations = []

        

    def normalize_bounds(self, bounds):
        """Makes sure all bounds feeded into GPyOpt are scaled to the domain [0, 1],
        to aid interpretation of convergence plots.

        Scalings are saved in instance parameters.

        Parameters
        ----------
        bounds : dicts
            Contains dicts with boundary information

        Returns
        -------
        scaled_bounds, scalings, intercepts : tuple
            Contains scaled bounds (list of dicts in GPy style), the scaling applied (numpy array)
            and an intercept (numpy array) to transform values back to their original domain
        """
        scaled_bounds = []
        scalings = []
        intercepts = []
        
        non_fixed_params = []
        
        print(self.device)
        
        for name, domain in self.bounds.items():
            # Get any fixed parmeters
            if type(domain) == int or type(domain) == float:
                # Take note
                self.fixed_parameters.append(name)

            # Free parameters
            elif type(domain) == tuple:
                # Bookkeeping
                self.free_parameters.append(name)

                # Get scaling
                lower_bound = min(domain)
                upper_bound = max(domain)
                scale = upper_bound - lower_bound

                # Transform to [0, 1] domain
                #scaled_bound = {'name': name, 'type': 'continuous', 'domain': (0., 1.)} #torch.adjustment required
                non_fixed_params.append(name)
                
                # Store
                #scaled_bounds.append(scaled_bound)
                scalings.append(scale)
                intercepts.append(lower_bound)
            else:
                raise ValueError("Domain bounds not understood")
        
        n_hyperparams = len(non_fixed_params)
        
        scaled_bounds = cat([zeros(1,n_hyperparams, device = self.device), 
                                   ones(1, n_hyperparams, device = self.device)], 0)
        return scaled_bounds, tensor(scalings, device = self.device, requires_grad = False), tensor(intercepts, device = self.device, requires_grad = False) #torch.adjustment required

    def denormalize_bounds(self, normalized_arguments):
        """Denormalize arguments to feed into model.

        Parameters
        ----------
        normalized_arguments : numpy array
            Contains arguments in same order as bounds

        Returns
        -------
        denormalized_arguments : 1-D numpy array
            Array with denormalized arguments

        """
        denormalized_bounds = (normalized_arguments * self.bound_scalings) + self.bound_intercepts
        return denormalized_bounds

    def construct_arguments(self, x):
        """Constructs arguments for ESN input from input array.

        Does so by denormalizing and adding arguments not involved in optimization,
        like the random seed.

        Parameters
        ----------
        x : 1-D numpy array
            Array containing normalized parameter values

        Returns
        -------
        arguments : dict
            Arguments that can be fed into an ESN

        """

        # Denormalize free parameters
        denormalized_values = self.denormalize_bounds(x)
        arguments = dict(zip(self.free_parameters, denormalized_values.flatten()))
        

        self.log_vars = ['connectivity', 'llambda', 'llambda2', 'enet_strength',
                         'noise', 'regularization', 'dt', 'gamma_cyclic' 
                         ]

        # Add fixed parameters
        for name in self.fixed_parameters:
            value = self.bounds[name]
            arguments[name] = value
            # if name in self.log_vars:
            #     arguments[name] = 10. ** value
            # else:
                

        for var in self.log_vars:
            if var in arguments:
                arguments[var] = 10. ** arguments[var]  # Log scale correction

        #assert False, f'args {arguments}'

        if 'n_nodes' in arguments:
            arguments['n_nodes'] = tensor(arguments['n_nodes'], dtype = torch.int32, device = self.device, requires_grad = False)  # Discretize #torch.adjustment required

        if not self.feedback is None:
            arguments['feedback'] = self.feedback
        
        for argument, val_tensor in arguments.items():
            
            try:
                arguments[argument] = arguments[argument].item()
            except:
                arguments[argument] = arguments[argument]
        return arguments

    def validate_data(self, y, x=None, verbose=True):
        """Validates inputted data against errors in shape and common mistakes.

        Parameters
        ----------
        y : numpy array
            A y-array to be checked (should be 2-D with series in columns)
        x : numpy array or None
            Optional x-array to be checked (should have same number of rows as y)
        verbose: bool
            Toggle to flag printed messages about common shape issues

        Raises
        ------
        ValueError
            Throws ValueError when data is not in the correct format.

        """
        # Check dimensions
        if not self.ODE_order:
            if not y.ndim == 2:
                raise ValueError("y-array is not 2 dimensional, if ODE and you didn't provide y then x is one dim")

            if verbose and y.shape[0] < y.shape[1]:
                print("Warning: y-array has more series (columns) than samples (rows). Check if this is correct")

        # Checks for x
        if self.ODE_order and x is None:
            assert False
        if not x is None:

            # Check dimensions
            if not x.ndim == 2:
                raise ValueError("x-array is not 2 dimensional")

            # Check shape equality
            if x.shape[0] != y.shape[0]:
                raise ValueError("y-array and x-array have different number of samples (rows)")
    
    # def eval_objective(self, x):
    #     """This is a helper function we use to unnormalize and evaluate a point"""
    #     return self.HRC(x) 
    #     #original BoTorch code:
    #     #unnormalize(x, self.scaled_bounds))

    def objective_function(self, parameters, train_y, validate_y, train_x=None, validate_x=None, random_seed=None):
        """Returns selected error metric on validation set.

        Parameters
        ----------
        parameters : array
            Parametrization of the Echo State Network, in column vector shape: (n, 1).
        train_y : array
            Dependent variable of the training set
        validate_y : array
            Dependent variable of the validation set
        train_x : array or None
            Independent variable(s) of the training set
        validate_x : array or None
            Independent variable(s) of the validation set

        Returns
        -------
        score : float
            Score on provided validation set

        """
        arguments = self.construct_arguments(self.range_bounds)

        # Build network 
        esn = self.model(**arguments, activation_f = self.activation_function,
                plot = False, model_type = self.model_type,
                input_weight_type = self.input_weight_type, already_normalized = already_normalized)
                #random_seed = self.random_seed) Distance_matrix = self.Distance_matrix)
                #bs_idx = self.obs_index, resp_idx = self.target_index, 

        # Train
        esn.fit(x=train_x, y=train_y, burn_in=self.esn_burn_in)

        # Validation score
        score = esn.test2(x=validate_x, y=validate_y, scoring_method=self.scoring_method, 
                            steps_ahead=self.steps_ahead, alpha=self.alpha)

        return score

    def define_tr_val(self, inputs):
        """Splits training sets (X, y) into train and validate sets, in order to later compute multiple samples of the objective function.

        Parameters
        ----------
        inputs : dict
            contains start_index and random_seed. 
                    (+) start index determines where to start the cross-validated sample
                    (+) random seed defines the reservoir's random seed.
        Returns
        ----------
        dict of new training and validation sets
        """
        ####
        #print("Hayden edit: parameters: " + str(parameters))
        #print("Hayden edit: fixed parameters: " + str(self.fixed_parameters))
        #print("Hayden edit: free parameters: " + str(self.free_parameters))
        ####
        
        start_index, random_seed = inputs["start_index"], inputs["random_seed"]
        train_stop_index = start_index + self.train_length
        validate_stop_index = train_stop_index + self.validate_length
        
        # Get samples
        if self.ODE_order:
            train_y = None
            validate_y = None
        else:
            train_y = self.y[start_index: train_stop_index]
            validate_y = self.y[train_stop_index: validate_stop_index]
            

        if not self.x is None:
            train_x = self.x[start_index: train_stop_index]
            validate_x = self.x[train_stop_index: validate_stop_index]
            
        else:
            train_x = None
            validate_x = None

        return {"tr_x" : train_x, "tr_y": train_y, "val_x": validate_x, "val_y" : validate_y }
        
        
    def build_unq_dict_lst(self, lst1, lst2, key1 = "start_index", key2 = "random_seed"):
        """This function builds a list of dictionairies with unique keys.

        Arguments: TODO
        """
        dict_lst = []
        for i in range(len(lst1)):
            for j in range(len(lst2)):
                dictt = {}
                dictt[key1] =  lst1[i]
                dictt[key2] =  lst2[j]
                dict_lst.append(dictt)
        return dict_lst

    def objective_sampler(self):
        """Splits training set into train and validate sets, and computes multiple samples of the objective function.

        This method also deals with dispatching multiple series to the objective function if there are multiple,
        and aggregates the returned scores by averaging.

        Parameters
        ----------
        parameters : array
            Parametrization of the Echo State Network

        Returns
        -------
        mean_score : 2-D array
            Column vector with mean score(s), as required by GPyOpt

        """
        # Get data
        #self.parameters = parameters
        training_y = self.y
        training_x = self.x

        
        # Set viable sample range
        if not self.ODE_order:
            #assert False
            viable_start = self.esn_burn_in
            # Get number of series
            self.n_series = training_y.shape[1]
            viable_stop = training_y.shape[0] - self.subsequence_length
        else:
            viable_start = 0
            # Get number of series
            self.n_series = training_x.shape[1]
            viable_stop = training_x.shape[0] - self.subsequence_length

        # Get sample lengths
        self.validate_length = torch.round(tensor(self.subsequence_length * self.validate_fraction, requires_grad  = False)).type(torch.int32)
        self.train_length = self.subsequence_length - self.validate_length

        ### TORCH
        start_indices = randint(low = viable_start, high = viable_stop, size = (self.cv_samples,))
        start_indices = [index_tensor.detach() for index_tensor in start_indices]
        
        if self.seed == None:
            random_seeds  = randint(0, 100000, size = (self.n_res,), generator = self.random_state) #device = self.device, 
        else:
            random_seeds = [self.seed]

        objective_inputs = self.build_unq_dict_lst(start_indices, random_seeds)

       
        """
        # Get samples
        if (self.cv_samples * self.n_res)  > 1:
            Pool = mp.Pool(self.cv_samples * self.n_res)

            #get the asynch object:
            results = list(zip(*mp.Pool.map(self.define_tr_val, objective_inputs)))

            mp.Pool.close()
            mp.Pool.join()
        else:
            results = self.define_tr_val(objective_inputs[0])

        self.scores = tensor(results).view(-1,1)

        #.reshape(scores.shape)

        mean_score = self.scores.mean()
        
        print('Score \u03BC:' + str(round(np.log10(mean_score),4)), ", \u03C3: ",  round(np.log10(self.scores.std()),4), "seed", random_seeds, "n", self.scores.shape[0])#, "scores", self.scores)#str(self.scores.std()),)
            # pars = self.construct_arguments(parameters)

        # Return scores
        return mean_score #.reshape(-1, 1)
        """
        return self.define_tr_val(objective_inputs[0])
    
    def my_loss_plot(self, ax, pred, start_loc, valid, steps_displated = 500):#, pred = pred):
        """

        Parameters
        ----------

        ax

        pred: ???
        ????

        start_loc: ???
        ????

        valid: ????
        ????

        steps displated: int
        
        Returns
        -------


        Arguments: TODO
        """
        pred_ = pred.cpu().numpy()

        ax.plot(range(len(valid)), valid, color = 'blue', label = "train")

        #ax.plot(valid, color = "green", label = "test", alpha = 0.4)
        for i in range(len(valid) - pred.shape[1]):
            if i % 2 == 0:
                ax.plot(range(i, pred_.shape[1]+ i + 1),
                         cat([valid[i], tensor(pred_[i,:])],0), color = "red", alpha = 0.3)
        #ax.set_xlim(start_loc, 2200)

        plt.legend()
    
    def train_plot_update(self, pred_, validate_y, steps_displayed, elastic_losses = None, restart_triggered = False):
        """If you are running rctorch in a jupyter notebook then this function displays live plots so that you can watch training if 
        self.interactive = True.

        

        Parameters
        ----------
        pred_: torch.tensor
            the model's prediction
        validate_y: torch.tensor
            the validation set of the response
        steps_displayed: int
            the number of timesteps of pred_ and validate_y to show in the plot
        elastic_losses:

        """
        if self.interactive:
            display.clear_output(wait=True) 
            
            pred_2plot = pred_.detach().to("cpu")
            if not self.ODE_order:
                validate_y_2plot = validate_y.detach().to("cpu")
            try:
                self.ax[1].clear()
                self.ax[0].clear()
            except:
                pass

            

            labels = "best value", "all samples"
            #log2 = np.log(2)

            # Plot 1: the training history of the bayesian optimization
            # if not self.log_score:
            #     self.ax[0].plot(np.log(self.errorz_step), alpha = 0.5, color = "blue", label = labels[0] )
            #     self.ax[0].plot(np.log(self.errorz), alpha = 0.2, color = "green", label = labels[1])
            # else:
            font_dict = {"prop" : {"size":14}}
            ticks_font_size = 14

            self.ax[0].plot(10**np.array(self.errorz_step), alpha = 0.5, color = "blue", label = labels[0] )
            self.ax[0].plot(10**np.array(self.errorz), alpha = 0.2, color = "green", label = labels[1])
            #self.ax[0].set_title("log(error) vs Thompson Sampling step")
            self.ax[0].set_ylabel(f"log({self.scoring_method})")
            self.ax[0].set_xlabel("BO step")
            self.ax[0].set_ylabel("Error")
            self.ax[0].legend(**font_dict)
            self.ax[0].set_yscale("log")
            #self.ax[0].set_ylim(10**-8,1)
            # self.ax[0].set_xtickslabels(fontsize= ticks_font_size )
            # self.ax[0].set_ytickslabels(fontsize= ticks_font_size )
            
            #plot 2: the turbo state            #plot 2: the turbo state
            self.ax[1].axhline(np.log(self.state.length_max)/self.log2, color = 'green', label = 'max length')
            self.ax[1].set_title("TURBO state")
            self.ax[1].plot(np.log(self.length_progress)/self.log2, color = 'blue', label = 'current length')
            self.ax[1].axhline(np.log(self.state.length_min)/self.log2, color = 'red', label = 'target length')
            self.ax[1].legend()  

            #plot 3 (most recent prediction)
            self.ax[2].clear()
            if not self.ODE_order:
                self.ax[2].plot(validate_y_2plot.to("cpu"), alpha = 0.5, color = "blue", label = "val set") #[:steps_displayed]
                self.ax[2].set_ylim(self.y.min().item() - 0.1, self.y.max().item() )
            self.ax[2].plot(pred_2plot[:steps_displayed], alpha = 0.3, color = "red", label = "latest pred")
                
            self.ax[2].set_title("Val Set Prediction")
            self.ax[2].set_ylabel("y")
            self.ax[2].set_xlabel(r'$t$')

            pl.legend()
            pl.tight_layout()

            display.display(pl.gcf())

            #clear the plot outputt and then re-plot
             

    def eval_objective(self, parameters, plot_type = "error", *args):
        """
        This version of the RC helper function

        Parameters
        -------
        parameters: torch.tensor
            a torch.tensor of the hyper-paramters drawn from the BO_step at time t

        plot_type

        Returns
        -------

        """
        # parameter_lst = []
        # for i in range(parameters.shape[0]):
        #     parameter_lst.append(self.construct_arguments(parameters[i, :]))

        parameter_lst = [self.construct_arguments(parameters[i, :]) for i in  range(parameters.shape[0])]

        #assert False, 'param_list ' + str(parameter_lst)
        
        start = time.time()

        #the algorithm is O(n) w.r.t. cv_samples.
        cv_samples = [self.objective_sampler() for i in range(self.cv_samples)]
        
        data_args = [(cv_samples, self.parallel_arguments, params, self.windowsOS, i) for i, params in enumerate(parameter_lst)]

        num_processes = parameters.shape[0]

        if self.batch_size > 1:
            Pool = mp_Pool(num_processes)

            #get the asynch object:
            results = Pool.map(execute_objective, data_args)
            
            Pool.close()
            Pool.join()
            results = sorted(results, key=lambda x: x[2]) 
            #try:
            Pool = mp_Pool(num_processes)

            #get the asynch object:
            results = Pool.map(execute_objective, data_args)
            
            Pool.close()
            Pool.join()
            results = sorted(results, key=lambda x: x[2]) 
            # except:
            #     #assert False
            #     results = [execute_HRC((cv_samples, parallel_arguments, params, self.windowsOS, i)) for i, params in enumerate(parameter_lst)]

            results = [(result[0], result[1]) for result in results]
            scores, preds = list(zip(*results)) 
        else:
            results = execute_objective(data_args[0])
            scores, preds, id_ = results
            scores, preds = [scores], [preds]

        if self.device == 'cuda':
            torch.cuda.empty_cache()
        
        for i, score in enumerate(scores):
            #score = score.view(1,1)
            if not i:
                Scores_ = [score]
            else:
                Scores_.append(score)
            self.errorz.append(score)
            #self.errorz_step.append(min(self.errorz))
            self.length_progress.append(self.state.length)

            
        #if self.count % self.batch_size == 0:
        if self.interactive : #and preds[0]["id_"] == 1
            self.train_plot_update(pred_ = preds[0]["pred"], validate_y = preds[0]["val_y"], 
                steps_displayed = preds[0]["pred"].shape[0]) #l2_prop  = self.l2_prop) #elastic_losses = RC.losses, 

        Scores_ = tensor(Scores_, dtype = self.dtype, device = self.device, requires_grad = False).unsqueeze(-1)

        #print('success_tolerance', self.state.success_counter)
        #print('failure_tolerance', self.state.failure_counter)
        #score_str = 'iter ' + str(self.count) +': Score ' + f'{error.type(torch.float64).mean():.4f}'# +', log(\u03BC):' + f'{log_mu:.4f}' 

        """
        score_str += " seed " + str(random_seeds) + " n " + str(self.scores.shape[0])

        if not self.count_:
            self.count_ = 1
        else:
            self.count_ +=1
        if log_mu > -0.5:
            printc(score_str, "fail")
        elif log_mu  > -0.8:
            printc(score_str, "warning")
        elif log_mu > -1:
            print(score_str)
        elif log_mu > -1.5:
            printc(score_str, "blue")
        elif log_mu > -2 :
            printc(score_str, "cyan")
        elif log_mu > -2.5:
            printc(score_str, "green")
        """
        stop = time.time()
        self.iteration_durations.append(stop - start)
        self.count += 1

        return - Scores_
    
    
    
    def optimize(self, y = None, x=None, store_path=None, epochs = 25, learning_rate = 0.005, scoring_method = "mse", criterion = MSELoss(), 
                    reparam_f = None, ODE_criterion = None, init_conditions = None, scale = True, force = None, backprop_f = None, backprop = False,
                     ode_coefs = None, solve = True, rounds = None, tr_score_prop = 0.5, q = None, eq_system = False, n_outputs = None, 
                     nonlinear_ode = False, reg_type = "nl_ham"):
        """Performs optimization (with cross-validation).

        Uses Bayesian Optimization with Gaussian Process priors to optimize ESN hyperparameters.

        Parameters
        ----------
        y : numpy array
            Column vector with target values (y-values)

        x : numpy array or None
            Optional array with input values (x-values)

        store_path : str or None
            Optional path where to store best found parameters to disk (in JSON)

        Returns
        -------
        best_arguments : numpy array
            The best parameters found during optimization

        """
        font = {'size'   : 18}
        plt.rc('font', **font)
        
        #self.multiple_ICs = True if len(init_conditions[0]) > 1 else False
        self.n_outputs = n_outputs
        if n_outputs != len(init_conditions):
            assert False, "n_outputs must match the len of ode_coefs and init_conds"
        self.nl = nonlinear_ode
        self.log2 = np.log(2)
        self.q = q
        self.eq_system = eq_system
        self.rounds = rounds
        self.tr_score_prop = tr_score_prop
        self.solve = solve
        self.ode_coefs = ode_coefs
        if self.batch_size > 1:
            if reg_type == "driven_pop":
                custom_loss = driven_pop_loss
                force = fforce
            elif reg_type == "simple_pop":
                custom_loss = driven_pop_loss
                force = no_fforce
            elif reg_type == "ham":
                custom_loss = ham_loss
            elif reg_type == "no_reg":
                custom_loss = no_reg_loss
            elif reg_type == "elastic":
                custom_loss = elastic_loss
            elif reg_type == "hennon":
                custom_loss = hennon_hailes_loss
            else:
                assert False
            self.ODE_criterion = custom_loss
            if backprop_f:
                self.backprop_f = optimize_last_layer
            else:
                self.backprop_f = None
            self.epochs = epochs
        else:
            self.ODE_criterion = ODE_criterion
            self.backprop_f = backprop_f

        self.init_conditions = init_conditions
        
        self.scale = scale

        if self.batch_size > 1:
            self.reparam_f = freparam
            #self.force = fforce
        else:
            self.reparam_f = reparam_f
        self.force = force

        self.scoring_method = scoring_method
        self.criterion = criterion
        self.epochs = epochs
        self.learning_rate = learning_rate

        """
        if self.ODE: 
            if self.dt != None:
                #self.alpha = self.leaking_rate[0] / self.dt

                start, stop = float(x[0]), float(x[-1])
                nsteps = int((stop - start) / self.dt)
                x = torch.linspace(start, stop, steps = nsteps, requires_grad=False).view(-1,1).to(self.device)
            elif type(x) == type([]) and len(X) == 3:
                x0, xf, nsteps = x #6*np.pi, 100
                x = torch.linspace(x0, xf, steps = nsteps, requires_grad=False).view(-1,1).to(self.device)
            else:
                assert False, "Please input start, stop, dt"
        else:
            #ensure that X is a two dimensional tensor, or if X is None declare a tensor.
            X = check_x(X, y, self.dev).to(self.device)
        """
        
        # Checks
        if not self.ODE_order:
            self.validate_data(y, x, self.verbose)
        
        # Initialize new random state
        if self.reservoir_matrices != None:
            if self.ODE_order:
                
                
                self.reservoir_matrices.n_inputs_ = x.shape[1]

            else:

                self.reservoir_matrices.n_inputs_ = max(y.shape[1] - 1, 1) if type(x) == type(None) else x.shape[1]
            
            self.reservoir_matrices.gen_in_weights()

        self.random_state = Generator().manual_seed(self.seed + 2)

        init_device = self.device if not self.windowsOS else torch.device('cpu')

        if not self.ODE_order:
            if type(y) == np.ndarray:
                y = torch.tensor(y, device = init_device, requires_grad = False)
            if len(y.shape) == 1:
                y = y.view(-1, 1)
            if y.device != self.device:
                y = y.to(init_device)
            self.y = y.type(self.dtype)  
        else:
            self.y = None

        self.x = x.type(self.dtype) if x is not None else None #torch.ones(*y.shape)

        # Inform user
        if self.verbose:
            print("Model initialization and exploration run...")
            
        if self.interactive:
            self.fig, self.ax = pl.subplots(1,3, figsize = (16,4))
            
        self.errorz, self.errorz_step, self.length_progress = [], [], []
        dim = len(self.free_parameters)
        self.state = TurboState(dim, length_min = self.length_min, 
                                batch_size=self.batch_size, success_tolerance = self.success_tolerance)
        
        self.count = 1


        declaration_args = {'activation_f' : self.activation_function,
                            'act_f_prime' : self.act_f_prime,
                                 #'backprop' : self.backprop,
                            'n_outputs' : self.n_outputs,
                                 #'model_type' : self.model_type,
                                 #'input_weight_type' : self.input_weight_type, 
                            'approximate_reservoir' : self.approximate_reservoir,
                            "device" : self.device,
                            "reservoir" : self.reservoir_matrices
                            }
        train_args = {"burn_in" : self.esn_burn_in, 
                       "ODE_order" : self.ODE_order,
                       #"track_in_grad" : self.track_in_grad,
                       "force" : self.force,
                       "reparam_f" : self.reparam_f,
                       "init_conditions" : self.init_conditions,
                       "ode_coefs" : self.ode_coefs,
                       "q" : self.q,
                       "eq_system" : self.eq_system,
                       "nl" : self.nl,
                       "backprop_f" : self.backprop_f,
                       "epochs" : self.epochs,
                       "SOLVE" : self.solve, 
                      #"track_in_grad" : False,
                      "init_conditions" : self.init_conditions,
                      #"SCALE" : self.scale,
                      "reparam_f" : self.reparam_f,
                      "ODE_criterion" : self.ODE_criterion
                       #"multiple_ICs" : self.multiple_ICs
                       }

        backprop_args = {"backprop_f" : self.backprop_f,
                         "epochs" : self.epochs}

        test_args = {"scoring_method" : self.scoring_method,
                     "reparam": self.reparam_f,
                     "ODE_criterion" : self.ODE_criterion}
        cv_args = {"rounds" : self.rounds,
                   "tr_score_prop" : self.tr_score_prop,
                   "log_score": self.log_score}

        self.parallel_arguments = {"declaration_args": declaration_args, #"RC" : RC,
                                   "train_args": train_args,
                                   "test_args" : test_args,
                                   "backprop_args" : backprop_args,
                                   "device" : self.device,
                                   "cv_args" : cv_args
                              }

        X_init = get_initial_points(self.scaled_bounds.shape[1], self.initial_samples, device = self.device, dtype = self.dtype)
        
        if len(X_init) > self.batch_size:
            for i in range(X_init.shape[0] // self.batch_size):
                #print(i)
                X_batch = X_init[ (i*self.batch_size) : ((i+1)*self.batch_size), : ]
                Y_batch = self.eval_objective( X_batch ) 
                
                if not i:
                    X_turbo = X_batch
                    Y_turbo = Y_batch 
                else:
                    Y_turbo = cat((Y_turbo, Y_batch), dim=0)
                    X_turbo = cat((X_turbo, X_batch), dim=0)
        else:

            X_turbo = X_init
            Y_turbo = self.eval_objective( X_init)
            
            #X_turbo.share_memory()
            #Y_turbo = tensor(
            #    [self.eval_objective(x.view(1,-1)) for x in X_turbo], dtype=dtype, device=device).unsqueeze(-1)
        X_turbo = X_turbo.to(self.device)
        Y_turbo = Y_turbo.to(self.device)

        self.X_turbo, self.Y_turbo =  X_turbo, Y_turbo
        
        n_init = self.initial_samples

        #append the errorz to errorz_step
        self.errorz_step += [max(self.errorz)] * X_turbo.shape[0] #n_init
        
        self.count = 0
        # Run until TuRBO converges
        while not self.state.restart_triggered:  

            # Fit a GP model
            train_Y = (Y_turbo - Y_turbo.mean()) / Y_turbo.std()
            likelihood = GaussianLikelihood(noise_constraint=Interval(1e-8, 1e-3))
            model = SingleTaskGP(X_turbo, train_Y, likelihood=likelihood)
            mll = ExactMarginalLogLikelihood(model.likelihood, model)
            fit_gpytorch_model(mll)

            # Create a batch
            X_next = generate_batch(
                state=self.state,
                model=model,
                X=X_turbo,
                Y=train_Y,
                batch_size=self.batch_size,
                n_candidates=min(5000, max(2000, 200 * dim)),
                num_restarts=10,
                raw_samples=512,
                acqf="ts",
                device = self.device
            )
            X_next = X_next

            #assert 1 ==0, X_next

            #can be parallelized:
            Y_next = self.eval_objective( X_next) #tensor()#.unsqueeze(-1)
                                    #[self.eval_objective(x.view(1,-1)) for x in X_next],
            print('Y_next', Y_next)
            print("self.state", self.state)
            # Update state 
            self.state = update_state(state=self.state, Y_next=Y_next)

            # Append data
            X_turbo = cat((X_turbo, X_next), dim=0)
            Y_turbo = cat((Y_turbo, Y_next), dim=0)
            
            # Print current status
            print( 
                f"{len(X_turbo)}) Best score: {max(Y_next).item():.4f},  TR length: {self.state.length:.2e}" + 
                f" length {self.state.length}"# Best value:.item() {state.best_value:.2e},
            )
            
            print( 
                f"TR length: {self.state.length:.2e}," +  f" min length {self.state.length_min:.2e}"
                # + Best value:.item() {state.best_value:.2e},
            )

            self.errorz_step += [min(self.errorz)] * self.batch_size

            #assert len(self.errorz) == len(self.errorz_step), "err len: {}, err step: {}".format(len(self.errorz), len(self.errorz_step) )
        else:
            display.clear_output()

        
        #display.clear_output(wait=True) 
        #display.display(pl.gcf())
                    
        # Save to disk if desired
        if not store_path is None:
            with open(store_path, 'w+') as output_file:
                json.dump(best_arguments, output_file, indent=4)
        
        best_vals = X_turbo[torch.argmax(Y_turbo)]

        
        
        denormed_ = self.denormalize_bounds(best_vals)
        
        try:
            denormed_ = denormalize_bounds(best_vals)
        except:
            print("FAIL")

        best_vals = X_turbo[torch.argmax(Y_turbo)]

        #####Bad temporary code to change it back into a dictionary
        denormed_free_parameters = list(zip(self.free_parameters, denormed_))
        denormed_free_parameters = dict([ (item[0], item[1].item()) for item in denormed_free_parameters])

        best_hyper_parameters = denormed_free_parameters
        for fixed_parameter in self.fixed_parameters:
            best_hyper_parameters = {fixed_parameter : self.bounds[fixed_parameter], **best_hyper_parameters }

        #log_vars = ['connectivity', 'llambda', 'llambda2', 'noise', 'regularization', 'dt']
        for var in self.log_vars:
            if var in best_hyper_parameters:
                best_hyper_parameters[var] = 10. ** best_hyper_parameters[var] 
                
        # Return best parameters
        return best_hyper_parameters #X_turbo, Y_turbo, state, best_vals, denormed_ #best_arguments



import torch
import torch.nn as nn
from torch.nn import KLDivLoss

class custom_kl_loss(nn.Module):
    """
    An example custom loss function
    
    See this link for more information:
    https://pdf.co/blog/deep-learning-pytorch-custom-loss-function
    
    Parameters
    ----------
        temp: The temperature argument for softmax (will soften the probability distribution)
    
    Returns
    ----------
        penalized loss metric
    """

    def __init__(self, **kwargs):
        super().__init__()
        self.temp = kwargs.get('temp', 1)
        
    def Tsoftmax(self, x, log = True):
        """ Temperature softens the output empirical distribution
        #https://stackoverflow.com/questions/44081007/logsoftmax-stability
        
        Parameters
        ----------
        x: torch.FloatTensor
            the vector to apply softmax to
        
        log: boolean
            if True returns the log softmax which is more stable because it lacks division
            otherwise return the normal softmax.
            
        #log(softmax) = x - b - log(x-b)exp(b)
        
        If we set formula, this new equation has both overflow and underflow stability conditions.
        You can also check this link for a little more descriptions:
        
        https://stackoverflow.com/questions/44081007/logsoftmax-stability
        """
        if log:
            #b = max(x)/self.temp
            #The torch.max function returns a tuple: input_max, input_indexes
            
            b  = torch.max(x)
            
            first_term = (x - b)/self.temp
            second_term = torch.log(torch.sum(torch.exp(first_term)))
            softmax = first_term - second_term
        else:
            exp_x = torch.exp(x/self.temp)
            softmax = exp_x / torch.sum(exp_x)
        return softmax

    def soften(self, target, input_, ):
        """
        kl expectes log inputs and probability outputs,
        From the documentation:
        
        As with :class:`~torch.nn.NLLLoss`, the `input` given is expected to contain
        *log-probabilities* and is not restricted to a 2D Tensor.
        The targets are interpreted as *probabilities* by default, but could be considered
        as *log-probabilities* with :attr:`log_target` set to ``True``.

        """
        soft_input = self.Tsoftmax(input_.view(-1,), log = True)
        
        soft_target = self.Tsoftmax(target.view(-1,), log = True)
        return soft_target, soft_input

    def forward(self, target, predicted, _lambda = 1, upper_error_limit = 100000, min_kl = None,  **kwargs):
        """
        Parameters:
            target: The target sequenc3
            prediction: The prediction sequence
            _lambda: a hyper parameter controlling the kl_penalty

        """
        #assert False, f'target {target.shape} predicted {predicted.shape}'
        self.kl =KLDivLoss(log_target = True)

        self.target = target
        self.predicted =  predicted
        self.errors = self.target - self.predicted

        soft_target, soft_input = self.soften(target, predicted)
        
        min_kl = self.kl(soft_target, soft_input)
        
        loss2 = torch.exp(self.kl(soft_target, soft_input)) - min_kl
        
        loss1 =  torch.mean(torch.square(self.errors))/torch.square(self.target).mean()

        assert loss1 >= 0, loss1
        assert loss2 >= 0, loss1
        
        
        loss = loss1 + _lambda * loss2

        #print(f'Loss: {loss1 } {loss2} {loss}')
        
        return loss

def odeLoss(time, N, dN_dx):
    f = 1 - torch.exp(-time)
    loss = f * dN_dx + N - 1
    loss = torch.square(loss).mean()
    return loss





from setuptools_scm import integration
from setuptools_scm.utils import do
from setuptools_scm import git
import pytest
from datetime import date


@pytest.fixture
def wd(wd):
    wd('git init')
    wd('git config user.email test@example.com')
    wd('git config user.name "a test"')
    wd.add_command = 'git add .'
    wd.commit_command = 'git commit -m test-{reason}'
    return wd


def test_version_from_git(wd):
    assert wd.version == '0.1.dev0'

    wd.commit_testfile()
    assert wd.version.startswith('0.1.dev1+g')
    assert not wd.version.endswith('1-')

    wd('git tag v0.1')
    assert wd.version == '0.1'

    wd.write('test.txt', 'test2')
    assert wd.version.startswith('0.2.dev0+g')

    wd.commit_testfile()
    assert wd.version.startswith('0.2.dev1+g')

    wd('git tag version-0.2')
    assert wd.version.startswith('0.2')

    wd.commit_testfile()
    wd('git tag version-0.2.post210+gbe48adfpost3+g0cc25f2')
    assert wd.version.startswith('0.2')


@pytest.mark.issue(179)
def test_unicode_version_scheme(wd):
    scheme = b'guess-next-dev'.decode('ascii')
    assert wd.get_version(version_scheme=scheme)


@pytest.mark.issue(108)
@pytest.mark.issue(109)
def test_git_worktree(wd):
    wd.write('test.txt', 'test2')
    # untracked files dont change the state
    assert wd.version == '0.1.dev0'
    wd('git add test.txt')
    assert wd.version.startswith('0.1.dev0+d')


@pytest.mark.issue(86)
def test_git_dirty_notag(wd):
    wd.commit_testfile()
    wd.write('test.txt', 'test2')
    wd("git add test.txt")
    assert wd.version.startswith('0.1.dev1')
    today = date.today()
    # we are dirty, check for the tag
    assert today.strftime('.d%Y%m%d') in wd.version


@pytest.fixture
def shallow_wd(wd, tmpdir):
    wd.commit_testfile()
    wd.commit_testfile()
    wd.commit_testfile()
    target = tmpdir.join('wd_shallow')
    do(['git', 'clone', "file://%s" % wd.cwd, str(target,), '--depth=1'])
    return target


def test_git_parse_shallow_warns(shallow_wd, recwarn):
    git.parse(str(shallow_wd))
    msg = recwarn.pop()
    assert 'is shallow and may cause errors' in str(msg.message)


def test_git_parse_shallow_fail(shallow_wd):
    with pytest.raises(ValueError) as einfo:
        git.parse(str(shallow_wd), pre_parse=git.fail_on_shallow)

    assert 'git fetch' in str(einfo.value)


def test_git_shallow_autocorrect(shallow_wd, recwarn):
    git.parse(str(shallow_wd), pre_parse=git.fetch_on_shallow)
    msg = recwarn.pop()
    assert 'git fetch was used to rectify' in str(msg.message)
    git.parse(str(shallow_wd), pre_parse=git.fail_on_shallow)


def test_find_files_stop_at_root_git(wd):
    wd.commit_testfile()
    wd.cwd.ensure('project/setup.cfg')
    assert integration.find_files(str(wd.cwd/'project')) == []


@pytest.mark.issue(128)
def test_parse_no_worktree(tmpdir):
    ret = git.parse(str(tmpdir))
    assert ret is None


def test_alphanumeric_tags_match(wd):
    wd.commit_testfile()
    wd('git tag newstyle-development-started')
    assert wd.version.startswith('0.1.dev1+g')


def test_git_archive_export_ignore(wd):
    wd.write('test1.txt', 'test')
    wd.write('test2.txt', 'test')
    wd.write('.git/info/attributes',
             # Explicitly include test1.txt so that the test is not affected by
             # a potentially global gitattributes file on the test machine.
             '/test1.txt -export-ignore\n/test2.txt export-ignore')
    wd('git add test1.txt test2.txt')
    wd.commit()
    assert integration.find_files(str(wd.cwd)) == ['test1.txt']


@pytest.mark.issue(228)
def test_git_archive_subdirectory(wd):
    wd('mkdir foobar')
    wd.write('foobar/test1.txt', 'test')
    wd('git add foobar')
    wd.commit()
    assert integration.find_files(str(wd.cwd)) == ['foobar/test1.txt']
